from .base import StateFunc
from .pipe import make_partial_pipe, make_pipe


@make_partial_pipe
def predict(x, state: StateFunc):
    return state._predict(x)


@make_pipe
def summary(state: StateFunc):
    return state._summary()



import jax.numpy as jnp
from equinox import filter_jit, filter_vmap
from jax import lax
from jax.scipy.special import gammaln

from ..core import make_partial_pipe


@make_partial_pipe
def combination(k, n, dtype=jnp.int32):
    """Calculates the number of combinations of k elements from a set of n elements.

    Args:
        k (int or array): Number of elements to choose from the set.
        n (int): Size of the set.

    Returns:
        int or array: The number of combinations of k elements from a set of n elements.

    Example:
        >>> combination(2, 5)
        10
    """
    k = jnp.asarray(k, dtype=dtype)
    k = jnp.atleast_1d(k)
    return filter_vmap(_comb)(k, n)


@filter_jit
def _cal_comb(k, n):
    log_kfrac = gammaln(k + 1)
    log_nfrac = gammaln(n + 1)
    log_n_kfrac = gammaln(n - k + 1)
    comb = jnp.exp(log_nfrac - log_kfrac - log_n_kfrac)
    comb = jnp.round(comb)
    comb = jnp.asarray(comb, dtype=jnp.int32)
    return comb


@filter_jit
def _comb(k, n):
    if_illegal = jnp.where(jnp.logical_or(k > n, k < 0), 1, 0)

    def func1(nn, kk):
        return jnp.asarray(0, dtype=jnp.int32)

    def func2(nn, kk):
        return _cal_comb(nn, kk)

    return jnp.asarray(lax.cond(if_illegal, func1, func2, k, n))



from typing import Optional, Union

import jax.numpy as jnp
import jax.random as jrand
from equinox import filter_grad, filter_jit, filter_vmap
from jax._src.random import Shape
from jax.random import KeyArray
from jax.scipy.special import ndtr, ndtri
from jaxtyping import ArrayLike, Bool, Float

from ..core import make_partial_pipe


@filter_jit
def _pnorm(
    x: Union[Float, ArrayLike],
    mean: Union[Float, ArrayLike] = 0.0,
    sd: Union[Float, ArrayLike] = 1.0,
):
    scaled = (x - mean) / sd
    return ndtr(scaled)


@make_partial_pipe(name="pnorm")
def pnorm(
    q: Union[Float, ArrayLike],
    mean: Union[Float, ArrayLike] = 0.0,
    sd: Union[Float, ArrayLike] = 1.0,
    lower_tail=True,
    log_prob=False,
    dtype=jnp.float32,
) -> ArrayLike:
    """Calculate the cumulative distribution function (CDF) of the normal distribution.

    Args:
        q (Union[Float, ArrayLike]): The quantiles to calculate the CDF at.
        mean (Union[Float, ArrayLike], optional): The mean of the normal distribution. Defaults to 0.0.
        sd (Union[Float, ArrayLike], optional): The standard deviation of the normal distribution.
            Defaults to 1.0.
        lower_tail (bool, optional): If True, calculate the probability that x is less than or equal to the
            given quantile(s). If False, calculate the probability that x is greater than the given quantile(s).
            Defaults to True.
        log_prob (bool, optional): If True, return the log of the CDF instead of the actual value.
            Defaults to False.
        dtype (jnp.dtype, optional): The dtype of the output. Defaults to jnp.float32.

    Returns:
        ArrayLike: The CDF of the normal distribution evaluated at x.


    Examples:
        >>> pnorm(2.0)
        Array([0.97724986], dtype=float32)

        >>> pnorm([1.5, 2.0, 2.5], mean=2.0, sd=0.5, lower_tail=False)
        Array([0.8413447 , 0.5       , 0.15865529], dtype=float32)
    """
    q = jnp.asarray(q, dtype=dtype)
    q = jnp.atleast_1d(q)
    p = filter_vmap(_pnorm)(q, mean, sd)
    if not lower_tail:
        p = 1.0 - p
    if log_prob:
        p = jnp.log(p)
    return p


_dnorm = filter_jit(filter_grad(_pnorm))


@make_partial_pipe(name="dnorm")
def dnorm(
    x: Union[Float, ArrayLike],
    mean: Union[Float, ArrayLike] = 0.0,
    sd: Union[Float, ArrayLike] = 1.0,
    lower_tail=True,
    log_prob=False,
    dtype=jnp.float32,
) -> ArrayLike:
    """
    Probability density function (PDF) for Normal distribution.

    Args:
        x (Union[Float, ArrayLike]): The input value(s) at which to evaluate the PDF.
        mean (Union[Float, ArrayLike], optional): The mean of the normal distribution. Defaults to 0.0.
        sd (Union[Float, ArrayLike], optional): The standard deviation of the normal distribution. Defaults to 1.0.
        lower_tail (bool, optional): If True (default), returns the cumulative distribution function (CDF) from negative infinity up to x. Otherwise, returns the CDF from x to positive infinity.
        log_prob (bool, optional): If True, returns the log-probability instead of the probability.
        dtype (jnp.dtype, optional): The dtype of the output. Default is `jnp.float32`.

    Returns:
        ArrayLike: The probability density function evaluated at point(s) x.

    Example:
        >>> import jax.numpy as jnp
        >>> x = jnp.array([0.5, 1.0, -1.5])
        >>> dnorm(x)
        Array([0.35206532, 0.24197075, 0.12951761], dtype=float32)
    """
    x = jnp.asarray(x, dtype=dtype)
    x = jnp.atleast_1d(x)
    grads = filter_vmap(_dnorm)(x, mean, sd)
    if not lower_tail:
        grads = 1 - grads
    if log_prob:
        grads = jnp.log(grads)
    return grads


@filter_jit
def _qnorm(
    p: Union[Float, ArrayLike],
    mean: Union[Float, ArrayLike] = 0.0,
    sd: Union[Float, ArrayLike] = 1.0,
):
    x = ndtri(p)
    return x * sd + mean


@make_partial_pipe(name="qnorm")
def qnorm(
    p: Union[Float, ArrayLike],
    mean: Union[Float, ArrayLike] = 0.0,
    sd: Union[Float, ArrayLike] = 1.0,
    lower_tail=True,
    log_prob=False,
    dtype=jnp.float32,
) -> ArrayLike:
    """
    Calculates the quantile function of the normal distribution for a given probability.

    Args:
        p (float or jnp.ndarray): Probability values.
        mean (float or jnp.ndarray, optional): Mean of the normal distribution. Default is 0.0.
        sd (float or jnp.ndarray, optional): Standard deviation of the normal distribution. Default is 1.0.
        lower_tail (bool, optional): If `True`, returns P(X ≤ x). If `False`, returns P(X > x). Default is `True`.
        log_prob (bool, optional): If `True`, returns the logarithm of the quantile function. Default is `False`.
        dtype (jnp.dtype, optional): The dtype of the output. Default is `jnp.float32`.

    Returns:
        ArrayLike: The inverse cumulative density function of the normal distribution evaluated at `q`.


    Examples:
        >>> qnorm(0.5)
        Array([0.], dtype=float32)
        >>> qnorm([0.25, 0.75], mean=3, sd=2)
        Array([1.6510204, 4.3489795], dtype=float32)
    """
    p = jnp.asarray(p, dtype=dtype)
    p = jnp.atleast_1d(p)
    if not lower_tail:
        p = 1 - p
    if log_prob:
        p = jnp.exp(p)
    x = filter_vmap(_qnorm)(p, mean, sd)
    return x


@filter_jit
def _rnorm(key, mean, sd, sample_shape, dtype):
    if sample_shape is None:
        sample_shape = jnp.broadcast_shapes(jnp.shape(mean), jnp.shape(sd))
    mean = jnp.broadcast_to(mean, sample_shape)
    sd = jnp.broadcast_to(sd, sample_shape)
    return jrand.normal(key, sample_shape, dtype=dtype) * sd + mean


@make_partial_pipe(name="rnorm")
def rnorm(
    key: KeyArray,
    sample_shape: Optional[Shape] = None,
    mean: Union[Float, ArrayLike] = 0.0,
    sd: Union[Float, ArrayLike] = 1.0,
    lower_tail: Bool = True,
    log_prob: Bool = False,
    dtype=jnp.float32,
) -> ArrayLike:
    """Generates random variables from a normal distribution.

    Args:
        key: A KeyArray object used to generate the random numbers.
        sample_shape: An optional tuple of integers specifying the shape of the
        output array. Defaults to an empty tuple.
        mean: The mean of the normal distribution. Defaults to 0.0.
        sd: The standard deviation of the normal distribution. Defaults to 1.0.
        lower_tail: If True (default), returns the cumulative distribution function (CDF) from negative infinity up to x. Otherwise, returns the CDF from x to positive infinity.
        log_prob: If True, returns the log-probability instead of the probability.
        dtype: The dtype of the output. Defaults to jnp.float32.

    Returns:
        ArrayLike: Random samples from a normal distribution.

    Example:
        >>> import jax.numpy as jnp
        >>> from jax import random
        >>> key = random.PRNGKey(0)
        >>> rnorm(key, sample_shape=(3, 2))
        Array([[ 0.18784384, -1.2833426 ],
                [ 0.6494181 ,  1.2490594 ],
                [ 0.24447003, -0.11744965]], dtype=float32)

    """
    rvs = _rnorm(key, mean, sd, sample_shape, dtype=dtype)
    if not lower_tail:
        rvs = 1 - rvs

    if log_prob:
        rvs = jnp.log(rvs)

    return rvs



from typing import AnyStr

from jaxtyping import ArrayLike

from ..core import StateFunc


class HypoTest(StateFunc):
    statistic: ArrayLike
    parameters: ArrayLike
    p_value: ArrayLike
    estimate: ArrayLike
    null_value: ArrayLike
    alternative: AnyStr

    def __init__(
        self,
        statistic=None,
        parameters=None,
        p_value=None,
        estimate=None,
        null_value=None,
        alternative=None,
        name='HTest'
    ):
        super().__init__(name=name, fn=None)
        self.statistic = statistic
        self.parameters = parameters
        self.p_value = p_value
        self.estimate = estimate
        self.null_value = null_value
        self.alternative = alternative
    



"""Stamox: Stats Models in JAX Library"""

import sys

from stamox import (
    anova,
    basic,
    cluster,
    core,
    distribution,
    experimental,
    formula,
    hypothesis,
    math,
    regression,
    sample,
    transformation,
)


def _check_py_version():
    if sys.version_info[0] < 3:
        raise Exception("Please use Python 3. Python 2 is not supported.")


_check_py_version()


__all__ = [
    "anova",
    "basic",
    "cluster",
    "core",
    "distribution",
    "experimental",
    "formula",
    "hypothesis",
    "math",
    "regression",
    "sample",
    "transformation",
]



import jax.numpy as jnp
from jax import vmap
from jaxtyping import ArrayLike

from .core import make_partial_pipe


mean = make_partial_pipe(jnp.mean)
sd = make_partial_pipe(jnp.std)
var = make_partial_pipe(jnp.var)
median = make_partial_pipe(jnp.median)
quantile = make_partial_pipe(jnp.quantile)
min = make_partial_pipe(jnp.min)
max = make_partial_pipe(jnp.max)
sum = make_partial_pipe(jnp.sum)
prod = make_partial_pipe(jnp.prod)
cumsum = make_partial_pipe(jnp.cumsum)
cumprod = make_partial_pipe(jnp.cumprod)
diff = make_partial_pipe(jnp.diff)
cov = make_partial_pipe(jnp.cov)
corrcoef = make_partial_pipe(jnp.corrcoef)
arcsin = make_partial_pipe(jnp.arcsin)
arccos = make_partial_pipe(jnp.arccos)


@make_partial_pipe
def scale(x: ArrayLike, axis: int = 0) -> ArrayLike:
    """Calculate standardized x along axis.

    Args:
        x (array-like): Input array.
        axis (int, optional): Axis along which to calculate mean and standard deviation. Defaults to 0.

    Returns:
        ArrayLike: Standardized x along axis.
    """
    x = jnp.asarray(x)
    _mean = mean(x, axis=axis)
    _std = sd(x, axis=axis, ddof=1)
    _scaled = vmap(lambda a, b, c: (a - b) / c, in_axes=(axis, None, None))(
        x, _mean, _std
    )
    return _scaled



from stamox.core._func_state import predict, summary
from stamox.core.base import (
    Functional,
    StateFunc,
)
from stamox.core.jit import partial_pipe_jit, pipe_jit
from stamox.core.maps import partial_pipe_pmap, partial_pipe_vmap, pipe_pmap, pipe_vmap
from stamox.core.pipe import (
    make_partial_pipe,
    make_pipe,
    Pipe,
    Pipeable,
)


__all__ = [
    "Functional",
    "make_partial_pipe",
    "make_pipe",
    "partial_pipe_jit",
    "partial_pipe_pmap",
    "partial_pipe_vmap",
    "Pipe",
    "Pipeable",
    "pipe_jit",
    "pipe_pmap",
    "pipe_vmap",
    "predict",
    "StateFunc",
    "summary",
]



from typing import Optional, Union

import jax.numpy as jnp
import jax.random as jrand
from equinox import filter_jit, filter_vmap
from jax import lax
from jax._src.random import Shape
from jax.random import KeyArray
from jaxtyping import ArrayLike, Float, Int

from ..core import make_partial_pipe


@filter_jit
def _prademacher(k: Union[Int, Float, ArrayLike]):
    cond0 = k < -1.0
    cond1 = jnp.logical_and(k >= -1.0, k < 1.0)
    cond2 = k >= 1.0
    index = jnp.argwhere(jnp.array([cond0, cond1, cond2]), size=1).squeeze()
    branches = [lambda: 0.0, lambda: 1 / 2, lambda: 1.0]
    p = lax.switch(index, branches, k)
    return p


@make_partial_pipe
def prademacher(
    k: Union[Int, Float, ArrayLike], lower_tail: bool = True, log_p: bool = False
):
    k = jnp.atleast_1d(k)
    p = filter_vmap(_prademacher)(k)
    if not lower_tail:
        p = 1 - p
    if log_p:
        p = jnp.log(p)
    return p


@filter_jit
def _drademacher(k: Union[Int, Float, ArrayLike]):
    in_support = jnp.logical_or(k == -1.0, k == 1.0)
    dens = jnp.where(in_support, 1.0 / 2.0, 0)
    return dens


@make_partial_pipe
def drademacher(
    k: Union[Int, Float, ArrayLike], lower_tail: bool = True, log_prob: bool = False
):
    k = jnp.atleast_1d(k)
    grads = filter_vmap(_drademacher)(k)
    if not lower_tail:
        grads = 1 - grads
    if log_prob:
        grads = jnp.log(grads)
    return grads


@filter_jit
def _rrademacher(key: KeyArray, sample_shape: Optional[Shape] = None):
    return jrand.rademacher(key, shape=sample_shape)


@make_partial_pipe
def rrademacher(
    key: KeyArray,
    sample_shape: Optional[Shape] = None,
    lower_tail: bool = True,
    log_prob: bool = False,
):
    probs = _rrademacher(key, sample_shape=sample_shape)
    if not lower_tail:
        probs = 1 - probs
    if log_prob:
        probs = jnp.log(probs)
    return probs



from typing import List, Tuple, Union

import equinox as eqx
import jax.numpy as jnp
from jaxtyping import ArrayLike
from pandas import DataFrame

from ..core import make_partial_pipe
from ..distribution import pF, pt
from ..formula import get_design_matrices
from ._base import RegState


class OLSState(RegState):
    _X_names: List[str]
    _y_names: List[str]
    _SSE: float
    _SSR: float
    _SST: float
    _coef_std: ArrayLike

    def __init__(
        self,
        in_features,
        out_features,
        coefs,
        df_resid,
        df_model,
        n_obs,
        resid,
        fitted_values,
        rank,
        SSE,
        SSR,
        SST,
        coefs_std,
        X_names=None,
        y_names=None,
        dtype=jnp.float32,
        name="OLSResult",
    ):
        super().__init__(
            in_features,
            out_features,
            coefs,
            df_resid,
            df_model,
            n_obs,
            resid,
            fitted_values,
            rank,
            dtype=dtype,
            name=name,
        )
        self._X_names = X_names
        self._y_names = y_names
        self._SSE = SSE
        self._SSR = SSR
        self._SST = SST
        self._coef_std = coefs_std

    @property
    def X_names(self):
        return self._X_names

    @property
    def y_names(self):
        return self._y_names

    @property
    def SSE(self):
        return self._SSE

    @property
    def SSR(self):
        return self._SSR

    @property
    def SST(self):
        return self._SST

    @property
    def F_value(self):
        return (self.SSR / self._df_model) / (self.SSE / self._df_resid)

    @property
    def F_pvalue(self):
        return 1 - pF(self.F_value, self._df_model, self._df_resid, dtype=self.dtype)

    @property
    def R2(self):
        return self.SSR / self.SST

    @property
    def R2_adj(self):
        return 1 - (self.SSE / self._df_resid) / (self.SST / (self._n_obs - 1))

    @property
    def coef_std(self):
        return self._coef_std

    @property
    def StdErr(self):
        return self.coef_std * self.resid_stderr

    @property
    def t_values(self):
        return jnp.divide(self.coefs, self.StdErr)

    @property
    def t_pvalues(self):
        return (
            pt(
                jnp.abs(self.t_values),
                self.df_resid,
                lower_tail=False,
                dtype=self.dtype,
            )
            * 2
        )

    @property
    def resid_stderr(self):
        return jnp.sqrt(self.SSE / self.df_resid)


@make_partial_pipe
def lm(
    data: Union[List, Tuple, DataFrame, ArrayLike],
    formula=None,
    subset=None,
    weights=None,
    NA_action="drop",
    method="qr",
    dtype=jnp.float32,
) -> OLSState:
    """Fits a linear model using the given data and parameters.

    Args:
        data (Union[List, Tuple, DataFrame, ArrayLike]): The data to fit the linear model with.
        formula (str, optional): A formula for the linear model. Defaults to None.
        subset (list, optional): A list of indices to use as a subset of the data. Defaults to None.
        weights (array-like, optional): An array of weights to apply to the data. Defaults to None.
        NA_action (str, optional): The action to take when encountering missing values. Defaults to "drop".
        method (str, optional): The method to use for fitting the linear model. Defaults to "qr".
        dtype (jnp.float32, optional): The data type to use for the linear model. Defaults to jnp.float32.

    Returns:
        OLSState: The state of the fitted linear model.

    Example:
        >>> from stamox.regression import lm
        >>> import pandas as pd
        >>> import numpy as np
        >>> np.random.seed(42)
        >>> X = np.random.uniform(size=(1000, 3))
        >>> y = (
                3 * X[:, 0]
                + 2 * X[:, 1]
                - 7 * X[:, 2]
                + 1.
            )
        >>> data = pd.DataFrame(
                np.concatenate([X, y.reshape((-1, 1))], axis=1),
                columns=["x1", "x2", "x3", "y"],
            )
        >>> res = lm(data, "y ~ x1 + x2 + x3")
        >>> res.coefs
        Array([ 1., 3.,  2., -7.], dtype=float32)

    """
    if formula is None:
        if isinstance(data, DataFrame):
            raise ValueError("formula is required when data is DataFrame")
        elif isinstance(data, list):
            y = data[0]
            X = data[1]
        elif isinstance(data, tuple):
            y = data[0]
            X = data[1]
        elif isinstance(data, ArrayLike):
            y = data[:, 0]
            X = data[:, 1:]
        else:
            raise ValueError("data is not supported")
        X_names = None
        y_names = None
    else:
        matrices = get_design_matrices(data, formula, NA_action=NA_action, dtype=dtype)
        y = matrices.y
        X = matrices.X
        X_names = matrices.X_names
        y_names = matrices.y_names
    if subset is not None:
        y = y[subset, :]
        X = X[subset, :]
    if weights is not None:
        weights = jnp.asarray(weights).reshape(-1, 1)
        W = jnp.diag(weights.reshape(-1))
        y = jnp.multiply(y, weights)
        X = jnp.matmul(X.T, W).T
    return _fit_lm(X, y, method, X_names=X_names, y_names=y_names)


@eqx.filter_jit
def _fit_lm(X, y, method, X_names=None, y_names=None):
    num_samples = X.shape[0]
    in_features = X.shape[1]
    out_features = 1
    if len(jnp.shape(y)) == 2:
        if jnp.shape(y)[1] > 1:
            raise NotImplementedError("multiple output is not supported yet.")
    elif len(jnp.shape(y)) == 1:
        y = jnp.atleast_1d(y)
    else:
        raise NotImplementedError("batchable output is not supported yet.")
    if method == "inv":
        X_pinv = jnp.linalg.inv(X.T @ X) @ X.T
    elif method == "qr":
        Q, R = jnp.linalg.qr(X)
        R_inv = jnp.linalg.inv(R)
        X_pinv = R_inv @ Q.T
    elif method == "svd":
        U, S, Vt = jnp.linalg.svd(X)
        D = jnp.zeros((num_samples, in_features))
        D = D.at[:in_features, :in_features].set(jnp.diag(1 / S))
        X_pinv = Vt.T @ D.T @ U.T
    else:
        raise NotImplementedError("Not Implemented")
    coefs = X_pinv @ y
    _coefs_std = jnp.sqrt(jnp.diag(X_pinv @ X_pinv.T)).reshape(-1, 1)
    _fitted_values = X @ coefs
    _residuals = y - _fitted_values
    _df_resid = num_samples - in_features
    _df_model = in_features - 1
    _rank = jnp.linalg.matrix_rank(X)
    _SSE = jnp.sum(_residuals**2)
    _SST = jnp.sum((y - jnp.mean(y)) ** 2)
    _SSR = _SST - _SSE
    state = OLSState(
        in_features=in_features,
        out_features=out_features,
        coefs=coefs,
        df_resid=_df_resid,
        df_model=_df_model,
        n_obs=num_samples,
        resid=_residuals,
        fitted_values=_fitted_values,
        rank=_rank,
        SSE=_SSE,
        SSR=_SSR,
        SST=_SST,
        coefs_std=_coefs_std,
        dtype=X.dtype,
        X_names=X_names,
        y_names=y_names,
    )
    return state



import inspect
from typing import TypeVar


_T = TypeVar("_T")

Self = TypeVar("Self", bound="better_partial")


class better_partial:
    __slots__ = "func", "keywords", "__dict__", "__weakref__"

    def __new__(cls, func, /, *args, **keywords) -> Self:
        if not callable(func):
            raise TypeError("the first argument must be callable")

        if hasattr(func, "func"):
            keywords = {**func.keywords, **keywords}
            func = func.func

        self = super(better_partial, cls).__new__(cls)
        # get all args name
        args_name = inspect.getfullargspec(func).args
        self.args_name = args_name
        self.func = func
        unassign_argsname = [arg for arg in args_name if arg not in keywords]
        for i, arg in enumerate(unassign_argsname):
            if i < len(args):
                keywords[arg] = args[i]
        self.keywords = keywords

        return self

    def __call__(self, *args, **keywords) -> _T:
        keywords = {**self.keywords, **keywords}
        unassign_argsname = [arg for arg in self.args_name if arg not in keywords]
        for i, arg in enumerate(unassign_argsname):
            if i < len(args):
                keywords[arg] = args[i]
            else:
                raise TypeError(f"missing required argument {arg}")

        return self.func(**keywords)



from typing import Optional, Union

import jax.numpy as jnp
import jax.random as jrand
from equinox import filter_jit, filter_vmap
from jax import pure_callback, ShapeDtypeStruct
from jax._src.random import Shape
from jax.random import KeyArray
from jax.scipy.special import gammainc, gammaln
from jaxtyping import Array, ArrayLike, Bool, Float
from scipy.stats import poisson

from ..core import make_partial_pipe


@filter_jit
def _ppoisson(x: Union[Float, ArrayLike], rate: Union[Float, ArrayLike]) -> Array:
    k = jnp.floor(x) + 1.0
    return 1 - gammainc(k, rate)


@make_partial_pipe
def ppoisson(
    q: Union[Float, ArrayLike],
    rate: Union[Float, ArrayLike],
    lower_tail=True,
    log_prob=False,
    dtype=jnp.float32,
) -> ArrayLike:
    """Computes the cumulative distribution function of the Poisson distribution.

    Args:
        q (Union[Float, ArrayLike]): The value at which to evaluate the CDF.
        rate (Union[Float, ArrayLike]): The rate parameter of the Poisson distribution.
        lower_tail (bool, optional): Whether to compute the lower tail of the CDF. Defaults to True.
        log_prob (bool, optional): Whether to return the log probability. Defaults to False.
        dtype (jnp.dtype, optional): The dtype of the output. Defaults to jnp.float32.

    Returns:
        ArrayLike: The cumulative distribution function of the Poisson distribution evaluated at `q`.
    """
    q = jnp.asarray(q, dtype=dtype)
    q = jnp.atleast_1d(q)
    p = filter_vmap(_ppoisson)(q, rate)
    if not lower_tail:
        p = 1 - p
    if log_prob:
        p = jnp.log(p)
    return p


@filter_jit
def _dpoisson(x, rate):
    e = jnp.exp(-rate)
    numerator = rate**x
    denominator = jnp.exp(gammaln(x + 1))
    return (numerator / denominator) * e


@make_partial_pipe
def dpoisson(
    x: Union[Float, ArrayLike],
    rate: Union[Float, ArrayLike],
    lower_tail=True,
    log_prob=False,
    dtype=jnp.float32,
) -> ArrayLike:
    """Computes the probability density function of the Poisson distribution.

    Args:
        x (Union[Float, ArrayLike]): The value at which to evaluate the PDF.
        rate (Union[Float, ArrayLike]): The rate parameter of the Poisson distribution.
        lower_tail (bool, optional): Whether to compute the lower tail of the PDF. Defaults to True.
        log_prob (bool, optional): Whether to return the log probability. Defaults to False.
        dtype (jnp.dtype, optional): The dtype of the output. Defaults to jnp.float32.

    Returns:
        ArrayLike: The probability density function of the Poisson distribution evaluated at `x`.
    """
    x = jnp.asarray(x, dtype=dtype)
    x = jnp.atleast_1d(x)
    grads = filter_vmap(_dpoisson)(x, rate)
    if not lower_tail:
        grads = 1 - grads
    if log_prob:
        grads = jnp.log(grads)
    return grads


@filter_jit
def _rpoisson(
    key: KeyArray,
    rate: Union[Float, ArrayLike],
    sample_shape: Optional[Shape] = None,
    dtype=jnp.int32,
):
    return jrand.poisson(key, rate, shape=sample_shape, dtype=dtype)


@make_partial_pipe
def rpoisson(
    key: KeyArray,
    sample_shape: Optional[Shape] = None,
    rate: Union[Float, ArrayLike] = None,
    lower_tail=True,
    log_prob=False,
    dtype=jnp.int32,
) -> ArrayLike:
    """Generates samples from the Poisson distribution.

    Args:
        key (KeyArray): Random number generator state used for random sampling.
        rate (Union[Float, ArrayLike]): The rate parameter of the Poisson distribution.
        sample_shape (Optional[Shape], optional): Shape of the output array. Defaults to None.
        lower_tail (bool, optional): Whether to return the lower tail probability. Defaults to True.
        log_prob (bool, optional): Whether to return the log probability. Defaults to False.
        dtype (jnp.dtype, optional): The dtype of the output. Defaults to jnp.int32.

    Returns:
        ArrayLike: Samples from the Poisson distribution.
    """
    rvs = _rpoisson(key, rate, sample_shape=sample_shape, dtype=dtype)
    if not lower_tail:
        rvs = 1 - rvs
    if log_prob:
        rvs = jnp.log(rvs)
    return rvs


@filter_jit
def _qpoisson(q, rate, dtype):
    result_shape_type = ShapeDtypeStruct(jnp.shape(q), dtype)
    _scp_poisson_ppf = lambda q, rate: poisson(rate).ppf(q).astype(dtype)
    p = pure_callback(_scp_poisson_ppf, result_shape_type, q, rate)
    return p


@make_partial_pipe
def qpoisson(
    p: Union[Float, ArrayLike],
    rate: Union[Float, ArrayLike],
    lower_tail: Bool = True,
    log_prob: Bool = False,
    dtype=jnp.int32,
) -> ArrayLike:
    """Computes the quantile function of the Poisson distribution.

    Args:
        p (Union[Float, ArrayLike]): The probability at which to evaluate the quantile function.
        rate (Union[Float, ArrayLike]): The rate parameter of the Poisson distribution.
        lower_tail (bool, optional): Whether to compute the lower tail of the quantile function. Defaults to True.
        log_prob (bool, optional): Whether to return the log probability. Defaults to False.
        dtype (jnp.dtype, optional): The dtype of the output. Defaults to jnp.int32.

    Returns:
        ArrayLike: The quantile function of the Poisson distribution evaluated at `p`.
    """
    p = jnp.asarray(p)
    p = jnp.atleast_1d(p)
    if not lower_tail:
        p = 1 - p
    if log_prob:
        p = jnp.exp(p)
    q = filter_vmap(_qpoisson)(p, rate, dtype)
    return q



from typing import Optional, Union

import jax.numpy as jnp
import jax.random as jrand
from equinox import filter_grad, filter_jit, filter_vmap
from jax._src.random import KeyArray, Shape
from jaxtyping import ArrayLike, Float

from ..core import make_partial_pipe
from ..math.special import fdtr, fdtri


@filter_jit
def _pf(
    x: Union[Float, ArrayLike],
    dfn: Union[Float, ArrayLike],
    dfd: Union[Float, ArrayLike],
):
    return fdtr(dfn, dfd, x)


@make_partial_pipe
def pF(
    q: Union[Float, ArrayLike],
    dfn: Union[Float, ArrayLike],
    dfd: Union[Float, ArrayLike],
    lower_tail: bool = True,
    log_prob: bool = False,
    dtype=jnp.float32,
) -> ArrayLike:
    """Calculates the cumulative distribution function of the F-distribution.

    Args:
        q (Union[Float, ArrayLike]): The value at which to evaluate the cdf.
        dfn (Union[Float, ArrayLike]): The numerator degrees of freedom.
        dfd (Union[Float, ArrayLike]): The denominator degrees of freedom.
        lower_tail (bool, optional): If True (default), the lower tail probability is returned.
        log_prob (bool, optional): If True, the logarithm of the probability is returned.
        dtype (jnp.dtype, optional): The dtype of the output (default is float32).

    Returns:
        ArrayLike: The cumulative distribution function evaluated at `q`.

    Example:
        >>> pF(1.0, 1.0, 1.0)
        Array([0.5000001], dtype=float32, weak_type=True)
    """
    q = jnp.asarray(q, dtype=dtype)
    q = jnp.atleast_1d(q)
    p = filter_vmap(_pf)(q, dfn, dfd)
    if not lower_tail:
        p = 1 - p
    if log_prob:
        p = jnp.log(p)
    return p


_df = filter_jit(filter_grad(_pf))


@make_partial_pipe
def dF(
    x: Union[Float, ArrayLike],
    dfn: Union[Float, ArrayLike],
    dfd: Union[Float, ArrayLike],
    lower_tail: bool = True,
    log_prob: bool = False,
    dtype=jnp.float32,
) -> ArrayLike:
    """Calculates the gradient of the cumulative distribution function for a given x, dfn and dfd.

    Args:
        x (Union[Float, ArrayLike]): The value at which to calculate the gradient of the cumulative distribution function.
        dfn (Union[Float, ArrayLike]): The numerator degrees of freedom.
        dfd (Union[Float, ArrayLike]): The denominator degrees of freedom.
        lower_tail (bool, optional): Whether to calculate the lower tail of the cumulative distribution function. Defaults to True.
        log_prob (bool, optional): Whether to return the log probability. Defaults to False.
        dtype (jnp.dtype, optional): The dtype of the output. Defaults to float32.

    Returns:
        ArrayLike: The gradient of the cumulative distribution function.

    Example:
        >>> dF(1.0, 1.0, 1.0)
        Array([0.1591549], dtype=float32, weak_type=True)
    """
    x = jnp.asarray(x, dtype=dtype)
    x = jnp.atleast_1d(x)
    grads = filter_vmap(_df)(x, dfn, dfd)
    if not lower_tail:
        grads = 1 - grads
    if log_prob:
        grads = jnp.log(grads)
    return grads


@filter_jit
def _qf(
    q: Union[Float, ArrayLike],
    dfn: Union[Float, ArrayLike],
    dfd: Union[Float, ArrayLike],
):
    return fdtri(dfn, dfd, q)


@make_partial_pipe
def qF(
    p: Union[Float, ArrayLike],
    dfn: Union[Float, ArrayLike],
    dfd: Union[Float, ArrayLike],
    lower_tail: bool = True,
    log_prob: bool = False,
    dtype=jnp.float32,
) -> ArrayLike:
    """Calculates the quantile function of a given distribution.

    Args:
        p (Union[Float, ArrayLike]): The quantile to calculate.
        dfn (Union[Float, ArrayLike]): The degrees of freedom for the numerator.
        dfd (Union[Float, ArrayLike]): The degrees of freedom for the denominator.
        lower_tail (bool, optional): Whether to calculate the lower tail or not. Defaults to True.
        log_prob (bool, optional): Whether to calculate the log probability or not. Defaults to False.

    Returns:
        ArrayLike: The calculated quantile.

    Example:
        >>> qF(0.5, 1.0, 1.0)
        Array([0.99999714], dtype=float32)
    """
    p = jnp.asarray(p, dtype=dtype)
    p = jnp.atleast_1d(p)
    if not lower_tail:
        p = 1 - p
    if log_prob:
        p = jnp.exp(p)
    return filter_vmap(_qf)(p, dfn, dfd)


@filter_jit
def _rf(
    key: KeyArray,
    dfn: Union[Float, ArrayLike],
    dfd: Union[Float, ArrayLike],
    sample_shape: Optional[Shape] = None,
    dtype = jnp.float32,
):
    if sample_shape is None:
        sample_shape = jnp.broadcast_shapes(jnp.shape(dfn), jnp.shape(dfd))
    dfn = jnp.broadcast_to(dfn, sample_shape)
    dfd = jnp.broadcast_to(dfd, sample_shape)
    return jrand.f(key, dfn, dfd, shape=sample_shape, dtype=dtype)


@make_partial_pipe
def rF(
    key: KeyArray,
    sample_shape: Optional[Shape] = None,
    dfn: Union[Float, ArrayLike] = None,
    dfd: Union[Float, ArrayLike] = None,
    lower_tail: bool = True,
    log_prob: bool = False,
    dtype = jnp.float32,
):
    """Generate random variates from F-distribution.

    Args:
        key (KeyArray): Random key used for PRNG.
        sample_shape (Optional[Shape], optional): Shape of the samples to be drawn. Defaults to None.
        dfn (Union[Float, ArrayLike]): Degrees of freedom in numerator.
        dfd (Union[Float, ArrayLike]): Degrees of freedom in denominator.
        lower_tail (bool, optional): Whether to calculate the lower tail probability. Defaults to True.
        log_prob (bool, optional): Whether to return the log probability. Defaults to False.
        dtype (jnp.dtype, optional): The dtype of the output. Defaults to float32.

    Returns:
        ArrayLike : Random variates from F-distribution.

    Example:
        >>> rF(jax.random.PRNGKey(0), dfn=1.0, dfd=1.0)
        Array(40.787617, dtype=float32)

    """
    rvs = _rf(key, dfn, dfd, sample_shape, dtype=dtype)
    if not lower_tail:
        rvs = 1 - rvs
    if log_prob:
        rvs = jnp.log(rvs)
    return rvs



# ::: stamox/hypothesis/_bartlett.py ::: 1 of 1 (100%)

from functools import partial
from typing import Sequence

import jax.numpy as jnp
from equinox import filter_jit
from jax import vmap
from jaxtyping import ArrayLike

from ..distribution import pchisq
from ._base import HypoTest


class BartlettTest(HypoTest):
    """Class for performing a Bartlett's test.

    This class is used to perform a Bartlett's test, which tests the null hypothesis that all input samples are from populations with equal variances.

    Attributes:
        statistic (float): The test statistic.
        parameters (int): The degrees of freedom.
        p_value (float): The p-value of the test.
    """

    def __init__(
        self,
        statistic=None,
        parameters=None,
        p_value=None,
        estimate=None,
        null_value=None,
        alternative=None,
    ):
        super().__init__(
            statistic,
            parameters,
            p_value,
            estimate,
            null_value,
            alternative,
            name="BartlettTest",
        )

    def __repr__(self):
        return f"{self.name}(statistic={self.statistic}, parameters={self.parameters}, p_value={self.p_value})"

    @property
    def df(self):
        return self.parameters


def bartlett_test_fun(*samples: Sequence[ArrayLike]) -> BartlettTest:
    """Calculates the Bartlett test statistic for multiple samples.

    Args:
        *samples Sequence[ArrayLike]): A sequence of 1-D arrays, each containing
            a sample of scores. All samples must have the same length.

    Returns:
        BartlettTest: The Bartlett Test object.

    Example:
        >>> bartlett_test([1, 2, 3], [1, 2, 3])
        BartlettTest(statistic=0.0, parameters=1, p_value=1.0)
    """
    samples = jnp.vstack(samples)
    return _bartlett(samples)


@filter_jit
def _bartlett(samples):
    k = samples.shape[0]
    Ni = jnp.asarray(vmap(jnp.size, in_axes=(0,))(samples), dtype=samples.dtype)
    ssq = vmap(partial(jnp.var, ddof=1), in_axes=(0,))(samples)
    Ntot = jnp.sum(Ni, axis=0)
    spsq = jnp.sum((Ni - 1) * ssq, axis=0) / (1.0 * (Ntot - k))
    numer = (Ntot * 1.0 - k) * jnp.log(spsq) - jnp.sum(
        (Ni - 1.0) * jnp.log(ssq), axis=0
    )
    denom = 1.0 + 1.0 / (3 * (k - 1)) * (
        (jnp.sum(1.0 / (Ni - 1.0), axis=0)) - 1.0 / (Ntot - k)
    )
    stats = numer / denom
    param = k - 1
    pval = pchisq(stats, param, lower_tail=False).squeeze()
    return BartlettTest(statistic=stats, parameters=param, p_value=pval)



from typing import Optional, Union

import jax.numpy as jnp
import jax.random as jrand
from equinox import filter_grad, filter_jit, filter_vmap
from jax._src.random import Shape
from jax.random import KeyArray
from jaxtyping import ArrayLike, Bool, Float

from ..core import make_partial_pipe


@filter_jit
def _pcauchy(
    x: Union[Float, ArrayLike],
    loc: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
) -> ArrayLike:
    scaled = (x - loc) / scale
    return jnp.arctan(scaled) / jnp.pi + 0.5


@make_partial_pipe
def pcauchy(
    q: Union[Float, ArrayLike],
    loc: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
    lower_tail: Bool = True,
    log_prob: Bool = False,
    dtype = jnp.float32,
) -> ArrayLike:
    """Calculates the cumulative denisty probability c function of the Cauchy distribution.

    Args:
        q (Union[Float, ArrayLike]): The value at which to evaluate the CDF.
        loc (Union[Float, ArrayLike], optional): The location parameter of the Cauchy distribution. Defaults to 0.0.
        scale (Union[Float, ArrayLike], optional): The scale parameter of the Cauchy distribution. Defaults to 1.0.
        lower_tail (Bool, optional): Whether to return the lower tail probability. Defaults to True.
        log_prob (Bool, optional): Whether to return the log probability. Defaults to False.
        dtype (jnp.dtype, optional): The dtype of the output. Defaults to jnp.float32.

    Returns:
        ArrayLike: The cumulative density function of the Cauchy distribution.

    Example:
        >>> pcauchy(1.0, loc=0.0, scale=1.0, lower_tail=True, log_prob=False)
        Array([0.75], dtype=float32, weak_type=True)
    """
    q = jnp.asarray(q, dtype=dtype)
    q = jnp.atleast_1d(q)
    p = filter_vmap(_pcauchy)(q, loc, scale)
    if not lower_tail:
        p = 1 - p
    if log_prob:
        p = jnp.log(p)
    return p


_dcauchy = filter_grad(filter_jit(_pcauchy))


@make_partial_pipe
def dcauchy(
    x: Union[Float, ArrayLike],
    loc: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
    lower_tail: Bool = True,
    log_prob: Bool = False,
    dtype = jnp.float32,
) -> ArrayLike:
    """Computes the pdf of the Cauchy distribution.

    Args:
        x (Union[Float, ArrayLike]): The input values.
        loc (Union[Float, ArrayLike], optional): The location parameter. Defaults to 0.0.
        scale (Union[Float, ArrayLike], optional): The scale parameter. Defaults to 1.0.
        lower_tail (Bool, optional): Whether to compute the lower tail. Defaults to True.
        log_prob (Bool, optional): Whether to compute the log probability. Defaults to False.
        dtype (jnp.dtype, optional): The dtype of the output. Defaults to jnp.float32.

    Returns:
        ArrayLike: The pdf of the Cauchy distribution.

    Example:
        >>> dcauchy(1.0, loc=0.0, scale=1.0, lower_tail=True, log_prob=False)
        Array([0.15915494], dtype=float32, weak_type=True)
    """
    x = jnp.asarray(x, dtype=dtype)
    x = jnp.atleast_1d(x)
    grads = filter_vmap(_dcauchy)(x, loc, scale)
    if not lower_tail:
        grads = 1 - grads
    if log_prob:
        grads = jnp.log(grads)
    return grads


@filter_jit
def _qcauchy(
    q: Union[Float, ArrayLike],
    loc: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
):
    return loc + scale * jnp.tan(jnp.pi * (q - 0.5))


@make_partial_pipe
def qcauchy(
    q: Union[Float, ArrayLike],
    loc: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
    lower_tail: Bool = True,
    log_prob: Bool = False,
    dtype = jnp.float32,
) -> ArrayLike:
    """Computes the quantile of the Cauchy distribution.

    Args:
        q (Union[float, array-like]): Quantiles to compute.
        loc (Union[float, array-like], optional): Location parameter. Defaults to 0.0.
        scale (Union[float, array-like], optional): Scale parameter. Defaults to 1.0.
        lower_tail (bool, optional): Whether to compute the lower tail. Defaults to True.
        log_prob (bool, optional): Whether to compute the log probability. Defaults to False.
        dtype (jnp.dtype, optional): The dtype of the output. Defaults to jnp.float32.

    Returns:
        ArrayLike: The quantiles of the Cauchy distribution.

    Example:
        >>> qcauchy(0.5, loc=1.0, scale=2.0, lower_tail=True, log_prob=False)
        Array([1.], dtype=float32, weak_type=True)
    """
    q = jnp.asarray(q, dtype=dtype)
    q = jnp.atleast_1d(q)
    if not lower_tail:
        q = 1 - q
    if log_prob:
        q = jnp.exp(q)
    return filter_vmap(_qcauchy)(q, loc, scale)


def _rcauchy(
    key: KeyArray,
    loc: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
    sample_shape: Optional[Shape] = None,
    dtype = jnp.float32,
):
    if sample_shape is None:
        sample_shape = jnp.broadcast_shapes(jnp.shape(loc), jnp.shape(scale))
    loc = jnp.broadcast_to(loc, sample_shape)
    scale = jnp.broadcast_to(scale, sample_shape)
    return jrand.cauchy(key, sample_shape, dtype=dtype) * scale + loc


@make_partial_pipe
def rcauchy(
    key: KeyArray,
    sample_shape: Optional[Shape] = None,
    loc: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
    lower_tail: Bool = True,
    log_prob: Bool = False,
    dtype = jnp.float32,
) -> ArrayLike:
    """Generates random samples from the Cauchy distribution.

    Args:
        key: A PRNGKey to use for generating the samples.
        sample_shape: The shape of the output array.
        loc: The location parameter of the Cauchy distribution.
        scale: The scale parameter of the Cauchy distribution.
        lower_tail: Whether to return the lower tail probability.
        log_prob: Whether to return the log probability.
        dtype: The dtype of the output.

    Returns:
        ArrayLike: An array of samples from the Cauchy distribution.

    Example:
        >>> key = jax.random.PRNGKey(0)
        >>> rcauchy(key, sample_shape=(2, 3), loc=0.0, scale=1.0)
        Array([[ 0.23841971, -3.0880406 ,  0.9507532 ],
                [ 2.8963416 ,  0.31303588, -0.14792857]], dtype=float32)
    """
    rvs = _rcauchy(key, loc, scale, sample_shape, dtype)
    if not lower_tail:
        rvs = 1 - rvs
    if log_prob:
        rvs = jnp.log(rvs)
    return rvs



from typing import Callable, TypeVar

import jax.random as jrandom
from equinox import filter_jit, filter_vmap
from jaxtyping import ArrayLike, PyTree

from ..core import make_partial_pipe


ReturnValue = TypeVar("ReturnValue")


@make_partial_pipe(name="bootstrap_sample")
def bootstrap_sample(
    data: ArrayLike, num_samples: int, *, key: jrandom.KeyArray = None
) -> ArrayLike:
    """Generates `num_samples` bootstrap samples from `data` with replacement.

    Args:
        data (array-like): The original data.
        num_samples (int): The number of bootstrap samples to generate.
        key (jrandom.KeyArray, optional): A random key array. Defaults to None.

    Returns:
        ArrayLike: An array of size `(num_samples, len(data))` containing the bootstrap samples.

    Example:
        >>> import jax.numpy as jnp
        >>> import jax.random as jrandom
        >>> from stamox.sample import bootstrap_sample
        >>> data = jnp.arange(10)
        >>> key = jrandom.PRNGKey(0)
        >>> bootstrap_sample(data, num_samples=3, key=key)
        Array([[9, 1, 6, 2, 9, 3, 9, 9, 4, 5],
                [4, 0, 4, 4, 6, 2, 5, 6, 5, 3],
                [7, 6, 9, 0, 0, 7, 0, 5, 8, 4]], dtype=int32)
    """
    # Determine the number of elements in the data
    n = data.shape[0]
    keys = jrandom.split(key, num_samples)

    @filter_jit
    def sample_fn(key: jrandom.KeyArray):
        # Draw n random indices from the data, with replacement
        sample_indices = jrandom.choice(key, n, (n,), replace=True)
        # Use the drawn indices to create a new bootstrap sample
        sample = data[sample_indices, ...]
        return sample

    samples = filter_vmap(sample_fn)(keys)
    return samples


@make_partial_pipe(name="bootstrap")
def bootstrap(
    data: ArrayLike,
    call: Callable[..., ReturnValue],
    num_samples: int,
    *,
    key: jrandom.KeyArray = None
) -> PyTree:
    """Generates `num_samples` bootstrap samples from `data` with replacement, and calls `call` on each sample.

    Args:
        data (array-like): The original data.
        call (Callable[..., ReturnValue]): The function to call on each bootstrap sample.
        num_samples (int): The number of bootstrap samples to generate.
        key (jrandom.KeyArray, optional): A random key array. Defaults to None.

    Returns:
        PyTree: The return value of `call` on each bootstrap sample.

    Example:
        >>> import jax.numpy as jnp
        >>> import jax.random as jrandom
        >>> from stamox.sample import bootstrap
        >>> data = jnp.arange(10)
        >>> bootstrap(data, jnp.mean, 3, key=key)
        Array([5.7000003, 3.9      , 4.6      ], dtype=float32)

    """
    samples = bootstrap_sample(data, num_samples=num_samples, key=key)
    return filter_jit(filter_vmap(lambda x: call(x)))(samples)



from typing import Callable, TypeVar

import jax.numpy as jnp
from equinox import filter_jit, filter_vmap
from jaxtyping import ArrayLike, PyTree

from ..core import make_partial_pipe


ReturnValue = TypeVar("ReturnValue")


def jackknife_sample_fun(data: ArrayLike) -> ArrayLike:
    """Generates `num_samples` jackknife samples from `data` with replacement.

    Args:
        data (array-like): The original data.

    Returns:
        ArrayLike: An array of size (len(data)-1, len(data)) containing the jackknife samples.

    Example:
        >>> import jax.numpy as jnp
        >>> from stamox.sample import jackknife_sample
        >>> data = jnp.arange(3)
        >>> jackknife_sample(data)
        Array([[1, 2],
                [0, 2],
                [0, 1]], dtype=int32)

    """
    n = jnp.shape(data)[0]

    @filter_jit
    def apply_except_one(x, i):
        idx = jnp.arange(x.shape[0])
        idx = jnp.where(idx >= i, idx + 1, idx)[:-1]
        return x[idx]

    samples = filter_vmap(apply_except_one, in_axes=(None, 0))(data, jnp.arange(n))

    return samples


@make_partial_pipe(name="jackknife")
def jackknife(data: ArrayLike, call: Callable[..., ReturnValue]) -> PyTree:
    """Computes the jackknife estimate of a given data set.

    Args:
        data (ArrayLike): The data set to be analyzed.
        call (Callable[..., ReturnValue]): A function to be applied to each sample.

    Returns:
        PyTree: The jackknife estimate of the data set.

    Example:
        >>> import jax.numpy as jnp
        >>> from stamox.sample import jackknife
        >>> data = jnp.arange(3)
        >>> jackknife(data, lambda x: jnp.mean(x))
        Array([1.5, 1. , 0.5], dtype=float32)
    """
    samples = jackknife_sample_fun(data)
    return filter_jit(filter_vmap(lambda x: call(x)))(samples)



import jax.numpy as jnp

from ..distribution import qnorm


def cor2ci_kendall(cor, n, ci=0.95, correction="fieller"):
    if correction == "fieller":
        tau_se = (0.437 / (n - 4)) ** 0.5
    else:
        tau_se = 1 / (n - 3) ** 0.5

    moe = qnorm(1 - (1 - ci) / 2) * tau_se

    zu = jnp.arctanh(cor) + moe
    zl = jnp.arctanh(cor) - moe

    ci_low = jnp.tanh(zl)
    ci_high = jnp.tanh(zu)

    return ci_low, ci_high


def cor2ci_spearman(cor, n, ci=0.95, correction="fieller"):
    if correction == "fieller":
        zrs_se = (1.06 / (n - 3)) ** 0.5
    elif correction == "bw":
        zrs_se = ((1 + cor ** 2 / 2) / (n - 3)) ** 0.5
    else:
        zrs_se = 1 / (n - 3) ** 0.5

    moe = qnorm(1 - (1 - ci) / 2) * zrs_se

    zu = jnp.arctanh(cor) + moe
    zl = jnp.arctanh(cor) - moe

    ci_low = jnp.tanh(zl)
    ci_high = jnp.tanh(zu)

    return ci_low, ci_high


def cor2ci_pearson(cor, n, ci=0.95, *args):
    z = jnp.arctanh(cor)
    se = 1 / jnp.sqrt(n - 3)

    alpha = 1 - (1 - ci) / 2

    ci_low = z - se * qnorm(alpha)
    ci_high = z + se * qnorm(alpha)

    ci_low = jnp.tanh(ci_low)
    ci_high = jnp.tanh(ci_high)

    return ci_low, ci_high



import jax.numpy as jnp
from jax.scipy.special import betainc
from tensorflow_probability.substrates.jax.math import special as tfp_special


def fdtrc(a, b, x):
    w = b / (b + a * x)
    return betainc(0.5 * b, 0.5 * a, w)


def fdtr(a, b, x):
    w = a * x
    w = w / (b + w)
    return betainc(0.5 * a, 0.5 * b, w)


def fdtri(a, b, y):
    y = 1.0 - y
    w = betainc(0.5 * b, 0.5 * a, 0.5)
    cond0 = (w > y) | (y < 0.001)
    w = jnp.where(
        cond0,
        tfp_special.betaincinv(0.5 * b, 0.5 * a, y),
        tfp_special.betaincinv(0.5 * a, 0.5 * b, 1.0 - y),
    )
    x = jnp.where(cond0, (b - b * w) / (a * w), b * w / (a * (1.0 - w)))
    return x



import jax.numpy as jnp
from jax import vmap, jit

from ..distribution import pnorm, pt


@jit
def cor2p(cor, n, method="pearson"):
    # Statistics
    if method == "kendall":
        stats = (3 * cor * jnp.sqrt(n * (n-1))) / jnp.sqrt(2 * (2 * n + 5))
    else:
        stats = cor * jnp.sqrt((n - 2) / (1 - cor ** 2))

    # p value
    if method == "kendall":
        p = 2 * pnorm(-jnp.abs(stats))
    else:
        p = 2 * pt(-jnp.abs(stats), df=n - 2)

    return stats, p

    



import jax.numpy as jnp
from jax import lax
from jaxtyping import ArrayLike

from .core import partial_pipe_jit


@partial_pipe_jit
def boxcox(x: ArrayLike, lmbda: ArrayLike) -> ArrayLike:
    """Computes the Box-Cox transformation of a given array.

    Args:
        x: An array-like object to be transformed.
        lmbda: An array-like object containing the lambda values for the transformation.

    Returns:
        The boxcox transformed array.

    Example:
        >>> from stamox.transformation import boxcox
        >>> import jax.numpy as jnp
        >>> x = jnp.array([1, 2, 3, 4, 5], dtype=jnp.float32)
        >>> lmbda = jnp.array([0, 0, 0, 0, 0], dtype=jnp.float32)
        >>> boxcox(x, lmbda)
        Array([0.        , 0.6931472 , 1.0986123 , 1.3862944 , 1.6094378 ], dtype=float32)
    """
    return lax.select(lmbda == 0, jnp.log(x), (x ** lmbda - 1) / lmbda)







import functools as ft

import jax.numpy as jnp
from jax import jit

from stamox.distribution import pbeta


def pearsonr(x, y, alternative='two-sided'):
    x = jnp.asarray(x)
    y = jnp.asarray(y)

    return _pearsonr(x, y, alternative)


@ft.partial(jit, static_argnames=('alternative', ))
def _pearsonr(x, y, alternative='two-sided'):
    n = x.size

    xm = x - x.mean()
    ym = y - y.mean()

    normxm = jnp.linalg.norm(xm)
    normym = jnp.linalg.norm(ym)

    r = jnp.dot(xm/normxm, ym/normym)

    r = jnp.maximum(jnp.minimum(r, 1.0), -1.0)
    ab = n / 2 - 1

    if alternative == 'two-sided':
        prob = 2 * pbeta(0.5 * (1 - jnp.abs(r)), ab, ab)
    elif alternative == 'less':
        prob = 1 - pbeta(0.5 * (1 - jnp.abs(r)), ab, ab)
    else:
        prob = pbeta(0.5*(1 - jnp.abs(r)), ab, ab)

    return r, prob



from typing import Callable, Optional, Tuple

import jax.numpy as jnp
from jax import lax
from jaxtyping import ArrayLike

from ._normal import qnorm


_EPSILON = 1e-15

_pf_n_  = 8    			
_pf_L_  = 2 			
_yLarge_ = 4096	
_incF_ = (1./64)	
_iShrink_ = 8   
_relTol_ = 1e-15			
_xf_ = 4

def do_search(Y_, z, p, incr, lower_tail, log_prob):
    cond1 = z >= p
    left = lax.select(lower_tail, cond1, ~cond1)
    if left:
        pass

    

def q_discr_check_boundary(x):
    return lax.select(x < 0., 0., x)


def q_discrete_body(p, mu, sigma, gamma, lower_tail, log_prob):
    z = qnorm(p, lower_tail=lower_tail, log_prob=log_prob)
    y = lax.round(mu + sigma * (z + gamma * (z * z - 1)))
    y = q_discr_check_boundary(y)

    if log_prob is True:
        e = _pf_L_ * _EPSILON
        if lower_tail is True:
            p = p * (1. + e)
        else:
            p = p * (1. - e)
    else:
        e = _pf_n_ * _EPSILON
        if lower_tail is True:
            p = p * (1 - e)
        else:
            p = lax.select(1 - p > _xf_ * e, p * (1 + e), p)
    






from typing import Optional, Union

import jax.numpy as jnp
import jax.random as jrand
from equinox import filter_grad, filter_jit, filter_vmap
from jax._src.random import KeyArray, Shape
from jaxtyping import ArrayLike, Float

from ..core import make_partial_pipe


@filter_jit
def _pexp(x: Union[float, ArrayLike], rate: float) -> Float:
    return -jnp.expm1(-rate * x)


@make_partial_pipe
def pexp(
    q: Union[Float, ArrayLike],
    rate: Union[Float, ArrayLike],
    lower_tail: bool = True,
    log_prob: bool = False,
    dtype=jnp.float32,
) -> ArrayLike:
    """Calculates the probability of a given value or array of values for an exponential distribution.

    Args:
        q: Union[Float, ArrayLike]. The value or array of values to calculate the probability of.
        rate: Union[Float, ArrayLike]. The rate parameter of the exponential distribution.
        lower_tail: bool, optional. Whether to return the lower tail probability (default is True).
        log_prob: bool, optional. Whether to return the log probability (default is False).
        dtype: jnp.dtype, optional. The dtype of the output (default is float32).

    Returns:
        ArrayLike: The probability of the given value or array of values.

    Example:
        >>> pexp(1.0, 0.5)
        Array([0.39346933], dtype=float32, weak_type=True)
    """
    q = jnp.asarray(q, dtype=dtype)
    q = jnp.atleast_1d(q)
    p = filter_vmap(_pexp)(q, rate)
    if not lower_tail:
        p = 1 - p
    if log_prob:
        p = jnp.log(p)
    return p


@filter_jit
def _qexp(q: Union[float, ArrayLike], rate: float) -> Float:
    return -jnp.log1p(-q) / rate


@make_partial_pipe
def qexp(
    p: Union[float, ArrayLike],
    rate: Union[float, ArrayLike],
    lower_tail=True,
    log_prob=False,
    dtype=jnp.float32,
) -> ArrayLike:
    """Computes the quantile of an exponential distribution.

    Args:
        p (Union[float, ArrayLike]): Probability or log probability.
        rate (Union[float, ArrayLike]): Rate parameter of the exponential distribution.
        lower_tail (bool, optional): Whether to compute the lower tail. Defaults to True.
        log_prob (bool, optional): Whether `p` is a log probability. Defaults to False.
        dtype (jnp.dtype, optional): The dtype of the output. Defaults to jnp.float32.

    Returns:
        ArrayLike: The quantile of the exponential distribution.

    Example:
        >>> qexp(0.5, 1.0)
        Array([0.6931472], dtype=float32, weak_type=True)
    """
    p = jnp.asarray(p, dtype=dtype)
    p = jnp.atleast_1d(p)
    if not lower_tail:
        p = 1 - p
    if log_prob:
        p = jnp.exp(p)
    x = filter_vmap(_qexp)(p, rate)
    return x


_dexp = filter_jit(filter_grad(_pexp))


@make_partial_pipe
def dexp(
    x: Union[Float, ArrayLike],
    rate: Union[Float, ArrayLike],
    lower_tail: bool = True,
    log_prob: bool = False,
    dtype=jnp.float32,
) -> ArrayLike:
    """Calculates the derivative of the exponential distribution.

    Args:
        x (Union[Float, ArrayLike]): The value at which to evaluate the derivative.
        rate (Union[Float, ArrayLike]): The rate parameter of the exponential distribution.
        lower_tail (bool, optional): Whether to calculate the lower tail probability. Defaults to True.
        log_prob (bool, optional): Whether to return the log probability. Defaults to False.
        dtype (jnp.dtype, optional): The dtype of the output. Defaults to jnp.float32.

    Returns:
        ArrayLike: The derivative of the exponential distribution evaluated at x.

    Example:
        >>> dexp(1.0, 0.5, lower_tail=True, log_prob=False)
        Array([0.30326533], dtype=float32, weak_type=True)
    """
    x = jnp.asarray(x, dtype=dtype)
    x = jnp.atleast_1d(x)
    grads = filter_vmap(_dexp)(x, rate)
    if not lower_tail:
        grads = -grads
    if log_prob:
        grads = jnp.log(grads)
    return grads


@filter_jit
def _rexp(
    key: KeyArray,
    rate: Union[Float, ArrayLike],
    sample_shape: Optional[Shape] = None,
    dtype=jnp.float32,
):
    if sample_shape is None:
        sample_shape = jnp.shape(rate)
    rate = jnp.broadcast_to(rate, sample_shape)
    return jrand.exponential(key, shape=sample_shape, dtype=dtype) / rate


@make_partial_pipe
def rexp(
    key: KeyArray,
    sample_shape: Optional[Shape] = None,
    rate: Union[Float, ArrayLike] = None,
    lower_tail: bool = True,
    log_prob: bool = False,
    dtype=jnp.float32,
) -> ArrayLike:
    """Generates random samples from the exponential distribution.

    Args:
        key (KeyArray): A PRNGKey to use for generating random numbers.
        sample_shape (Optional[Shape], optional): The shape of the output array. Defaults to None.
        rate (Union[Float, ArrayLike], optional): The rate parameter of the exponential distribution. Defaults to None.
        lower_tail (bool, optional): Whether to return the lower tail of the distribution. Defaults to True.
        log_prob (bool, optional): Whether to return the log probability of the samples. Defaults to False.
        dtype (jnp.dtype, optional): The dtype of the output. Defaults to jnp.float32.

    Returns:
        ArrayLike: An array of random samples from the exponential distribution.

    Example:
        >>> key = jax.random.PRNGKey(0)
        >>> rexp(key, sample_shape=(2, 3), rate=1.0, lower_tail=False, log_prob=True)
        Array([[-0.69314718, -0.69314718, -0.69314718],
               [-0.69314718, -0.69314718, -0.69314718]], dtype=float32)
    """
    rvs = _rexp(key, rate, sample_shape, dtype)
    if not lower_tail:
        rvs = 1 - rvs
    if log_prob:
        rvs = jnp.log(rvs)
    return rvs



"""
The Friedman test is a non-parametric statistical test developed by Milton Friedman.
Similar to the parametric repeated measures ANOVA, it is used to detect differences in treatments across multiple test attempts. 
The procedure involves ranking each row (or block) together, then considering the values of ranks by columns. 
Applicable to complete block designs, it is thus a special case of the Durbin test.
"""
from typing import Sequence

import jax.numpy as jnp
from equinox import filter_jit
from jax import vmap
from jax.scipy.stats import rankdata
from jaxtyping import ArrayLike

from ..distribution import pchisq
from ._base import HypoTest


class FriedmanTest(HypoTest):
    """Class for performing Friedman Rank Sum Test.

    This class is a subclass of HypoTest and provides methods to perform the 
    Friedman Rank Sum Test.

    Attributes:
        statistic (float): The test statistic.
        parameters (tuple): Parameters of the distribution.
        p_value (float): The p-value of the test.
    """
    def __init__(
        self,
        statistic=None,
        parameters=None,
        p_value=None,
        estimate=None,
        null_value=None,
        alternative=None,
        name="Friedman Rank Sum Test",
    ):
        super().__init__(
            statistic, parameters, p_value, estimate, null_value, alternative, name
        )
    
    def __repr__(self):
        return f"{self.name}(statistic={self.statistic}, parameters={self.parameters}, p_value={self.p_value})"


def friedman_test_fun(*samples: Sequence[ArrayLike]) -> FriedmanTest:
    """Computes the Friedman statistic for a set of samples.

    Args:
        *samples: A sequence of samples, each sample being a sequence of
            observations.

    Returns:
        FriedmanTest: The Friedman Test object.
    """
    ImportWarning("This function is not yet functioned with ties. Use with caution.")
    samples = jnp.vstack(samples).T
    return _friedman(samples)


@filter_jit
def _friedman(samples):
    """Computes the Friedman statistic for a set of samples.

    Args:
        samples: A 2D array of samples, with the first dimension representing
            the treatments and the second dimension representing the blocks.

    Returns:
        The computed Friedman statistic.
    """
    n_blocks, k_treatments = samples.shape
    ranks = vmap(lambda x: rankdata(x))(samples)
    avg_ranks = jnp.mean(ranks, axis=0)
    Q = 12.0 * n_blocks / (k_treatments * (k_treatments + 1)) * jnp.sum(
        avg_ranks**2, axis=0, keepdims=True
    ) - 3 * n_blocks * (k_treatments + 1)
    param = k_treatments - 1
    pval = pchisq(Q, param, lower_tail=False)
    return FriedmanTest(statistic=Q, parameters=param, p_value=pval)



from typing import Optional, Union

import jax.numpy as jnp
import jax.random as jrand
from equinox import filter_grad, filter_jit, filter_vmap
from jax._src.random import Shape
from jax.random import KeyArray
from jaxtyping import ArrayLike, Bool, Float

from ..core import make_partial_pipe


@filter_jit
def _plaplace(
    x: Union[Float, ArrayLike],
    loc: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
):
    scaled = (x - loc) / scale
    return 0.5 - 0.5 * jnp.sign(scaled) * jnp.expm1(-jnp.abs(scaled))


@make_partial_pipe
def plaplace(
    q: Union[Float, ArrayLike],
    loc: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
    lower_tail: Bool = True,
    log_prob: Bool = False,
    dtype=jnp.float32,
) -> ArrayLike:
    """Calculates the Laplace cumulative density function.

    Args:
        q (Union[Float, ArrayLike]): The value at which to evaluate the Plaplace PDF.
        loc (Union[Float, ArrayLike], optional): The location parameter of the Plaplace PDF. Defaults to 0.0.
        scale (Union[Float, ArrayLike], optional): The scale parameter of the Plaplace PDF. Defaults to 1.0.
        lower_tail (Bool, optional): Whether to return the lower tail of the Plaplace PDF. Defaults to True.
        log_prob (Bool, optional): Whether to return the logarithm of the Plaplace PDF. Defaults to False.

    Returns:
        ArrayLike: The Laplace CDF evaluated at `q`.

    Example:
        >>> plaplace(1.0, 1.0, 1.0)
        Array([0.5], dtype=float32, weak_type=True)
    """
    q = jnp.asarray(q, dtype=dtype)
    q = jnp.atleast_1d(q)
    p = filter_vmap(_plaplace)(q, loc, scale)
    if not lower_tail:
        p = 1 - p
    if log_prob:
        p = jnp.log(p)
    return p


_dlaplace = filter_grad(filter_jit(_plaplace))


@make_partial_pipe
def dlaplace(
    x: Union[Float, ArrayLike],
    loc: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
    lower_tail: Bool = True,
    log_prob: Bool = False,
    dtype=jnp.float32,
) -> ArrayLike:
    """Calculates the Laplace probability density function for a given x, location and scale.

    Args:
        x (Union[Float, ArrayLike]): The value at which to calculate the probability density.
        loc (Union[Float, ArrayLike], optional): The location parameter. Defaults to 0.0.
        scale (Union[Float, ArrayLike], optional): The scale parameter. Defaults to 1.0.
        lower_tail (Bool, optional): Whether to return the lower tail of the distribution. Defaults to True.
        log_prob (Bool, optional): Whether to return the logarithm of the probability. Defaults to False.
        dtype (jnp.dtype, optional): The dtype of the output. Defaults to jnp.float32.

    Returns:
        ArrayLike: The probability density at the given x.

    Example:
        >>> dlaplace(1.0, 1.0, 1.0)
        Array([0.], dtype=float32, weak_type=True)
    """
    x = jnp.asarray(x, dtype=dtype)
    x = jnp.atleast_1d(x)
    p = filter_vmap(_dlaplace)(x, loc, scale)
    if not lower_tail:
        p = -p
    if log_prob:
        p = jnp.log(p)
    return p


@filter_jit
def _qlaplace(
    q: Union[Float, ArrayLike],
    loc: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
):
    a = q - 0.5
    return loc - scale * jnp.sign(a) * jnp.log1p(-2 * jnp.abs(a))


@make_partial_pipe
def qlaplace(
    p: Union[Float, ArrayLike],
    loc: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
    lower_tail: Bool = True,
    log_prob: Bool = False,
    dtype=jnp.float32,
) -> ArrayLike:
    """Computes the quantile of the Laplace distribution.

    Args:
        p (Union[Float, ArrayLike]): Quantiles to compute.
        loc (Union[Float, ArrayLike], optional): Location parameter. Defaults to 0.0.
        scale (Union[Float, ArrayLike], optional): Scale parameter. Defaults to 1.0.
        lower_tail (Bool, optional): Whether to compute the lower tail. Defaults to True.
        log_prob (Bool, optional): Whether to compute the log probability. Defaults to False.
        dtype (jnp.dtype, optional): The dtype of the output. Defaults to jnp.float32.

    Returns:
        ArrayLike: The quantiles of the Laplace distribution.

    Example:
        >>> qlaplace(0.5, 1.0, 1.0)
        Array([1.], dtype=float32, weak_type=True)
    """
    p = jnp.asarray(p, dtype=dtype)
    p = jnp.atleast_1d(p)
    if not lower_tail:
        p = 1 - p
    if log_prob:
        p = jnp.exp(p)
    return filter_vmap(_qlaplace)(p, loc, scale)


@filter_jit
def _rlaplace(
    key: KeyArray,
    loc: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
    sample_shape: Optional[Shape] = None,
    dtype=jnp.float32,
):
    if sample_shape is None:
        sample_shape = jnp.broadcast_shapes(jnp.shape(loc), jnp.shape(scale))
    loc = jnp.broadcast_to(loc, sample_shape)
    scale = jnp.broadcast_to(scale, sample_shape)
    return jrand.laplace(key, sample_shape, dtype=dtype) * scale + loc


@make_partial_pipe
def rlaplace(
    key: KeyArray,
    sample_shape: Optional[Shape] = None,
    loc: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
    lower_tail: Bool = True,
    log_prob: Bool = False,
    dtype=jnp.float32,
) -> ArrayLike:
    """Generates random Laplace samples from a given key.

    Args:
        key (KeyArray): The PRNG key to use for generating the samples.
        sample_shape (Optional[Shape], optional): The shape of the output array. Defaults to None.
        loc (Union[Float, ArrayLike], optional): The location parameter of the Laplace distribution. Defaults to 0.0.
        scale (Union[Float, ArrayLike], optional): The scale parameter of the Laplace distribution. Defaults to 1.0.
        lower_tail (Bool, optional): Whether to return the lower tail probability. Defaults to True.
        log_prob (Bool, optional): Whether to return the log probability. Defaults to False.
        dtype (jnp.dtype, optional): The dtype of the output. Defaults to jnp.float32.

    Returns:
        ArrayLike: An array containing the random Laplace samples.

    Example:
        >>> rlaplace(key, (2, 3))
        Array([[-0.16134426,  1.6125823 , -0.6615164 ],
                [-1.5528525 , -0.21459664,  0.09816013]], dtype=float32)

    """
    rvs = _rlaplace(key, loc, scale, sample_shape, dtype=dtype)
    if not lower_tail:
        rvs = 1 - rvs
    if log_prob:
        rvs = jnp.log(rvs)
    return rvs



from typing import Optional

import jax.numpy as jnp
import numpy as np
from equinox import filter_jit, filter_vmap
from jax import pure_callback, ShapeDtypeStruct
from jax._src.random import KeyArray, Shape
from jaxtyping import ArrayLike, Bool
from scipy.stats import binom
from tensorflow_probability.substrates.jax.distributions import Binomial as tfp_Binomial

from ..core import make_partial_pipe


@filter_jit
def _pbinom(q, size, prob) -> ArrayLike:
    bino = tfp_Binomial(total_count=size, probs=prob)
    return bino.cdf(q)


@make_partial_pipe
def pbinom(
    q: ArrayLike,
    size: ArrayLike,
    prob: ArrayLike,
    lower_tail: Bool = True,
    log_prob: Bool = False,
    dtype=jnp.float32,
) -> ArrayLike:
    """Calculates the cumulative probability of a binomial distribution.

    Args:
        q (ArrayLike): The quantiles to compute.
        size (ArrayLike): The number of trials.
        prob (ArrayLike): The probability of success in each trial.
        lower_tail (Bool, optional): If True (default), the lower tail probability is returned.
        log_prob (Bool, optional): If True, the logarithm of the probability is returned.
        dtype (optional): The data type of the output array. Defaults to jnp.float32.

    Returns:
        ArrayLike: The cumulative probability of the binomial distribution.

    Example:
        >>> q = jnp.array([0.1, 0.5, 0.9])
        >>> size = 10
        >>> prob = 0.5
        >>> pbinom(q, size, prob)
    """
    q = jnp.asarray(q, dtype=dtype)
    q = jnp.atleast_1d(q)
    p = filter_vmap(_pbinom)(q, size, prob)
    if not lower_tail:
        p = 1 - p
    if log_prob:
        p = jnp.log(p)
    return p


@filter_jit
def _dbinom(q, size, prob) -> ArrayLike:
    bino = tfp_Binomial(total_count=size, probs=prob)
    return bino.prob(q)


@make_partial_pipe
def dbinom(
    q: ArrayLike,
    size: ArrayLike,
    prob: ArrayLike,
    lower_tail: Bool = True,
    log_prob: Bool = False,
    dtype=jnp.float32,
) -> ArrayLike:
    """Computes the probability of a binomial distribution.

    Args:
        q (ArrayLike): The value to compute the probability for.
        size (ArrayLike): The number of trials in the binomial distribution.
        prob (ArrayLike): The probability of success in each trial.
        lower_tail (Bool, optional): Whether to compute the lower tail probability. Defaults to True.
        log_prob (Bool, optional): Whether to return the logarithm of the probability. Defaults to False.
        dtype (jnp.float32, optional): The data type of the output array. Defaults to jnp.float32.

    Returns:
        ArrayLike: The probability of the binomial distribution.
    """
    q = jnp.asarray(q, dtype=dtype)
    q = jnp.atleast_1d(q)
    p = filter_vmap(_dbinom)(q, size, prob)
    if not lower_tail:
        p = 1 - p
    if log_prob:
        p = jnp.log(p)
    return p


@filter_jit
def _qbinom(p, size, prob, dtype) -> ArrayLike:
    result_shape_type = ShapeDtypeStruct(jnp.shape(p), dtype)
    _scp_binom_ppf = lambda x: binom(size, prob).ppf(x).astype(np.int32)
    q = pure_callback(_scp_binom_ppf, result_shape_type, p)
    return q


@make_partial_pipe
def qbinom(
    p: ArrayLike,
    size: ArrayLike,
    prob: ArrayLike,
    lower_tail: Bool = True,
    log_prob: Bool = False,
    dtype=jnp.int32,
) -> ArrayLike:
    """Computes the quantile of a binomial distribution.

    Args:
        p (ArrayLike): The probability of success.
        size (ArrayLike): The number of trials.
        prob (ArrayLike): The probability of success in each trial.
        lower_tail (Bool, optional): Whether to compute the lower tail or not. Defaults to True.
        log_prob (Bool, optional): Whether to compute the log probability or not. Defaults to False.
        dtype (jnp.int32, optional): The data type of the output array. Defaults to jnp.int32.

    Returns:
        ArrayLike: The quantile of the binomial distribution.
    """
    p = jnp.asarray(p)
    p = jnp.atleast_1d(p)
    if not lower_tail:
        p = 1 - p
    if log_prob:
        p = jnp.exp(p)

    q = filter_vmap(_qbinom)(p, size, prob, dtype)
    return q


@filter_jit
def _rbinom(key, n, prob, sample_shape, dtype) -> ArrayLike:
    bino = tfp_Binomial(total_count=n, probs=prob)
    return bino.sample(sample_shape=sample_shape, seed=key).astype(dtype)


@make_partial_pipe
def rbinom(
    key: KeyArray,
    sample_shape: Optional[Shape] = None,
    n: ArrayLike = None,
    prob: ArrayLike = None,
    lower_tail: Bool = True,
    log_prob: Bool = False,
    dtype=jnp.float32,
) -> ArrayLike:
    """Generates random binomial samples from a given probability distribution.

    Args:
        key (KeyArray): A random number generator key.
        sample_shape (Optional[Shape], optional): The shape of the output array. Defaults to None.
        n (ArrayLike, optional): The number of trials. Defaults to None.
        prob (ArrayLike, optional): The probability of success for each trial. Defaults to None.
        lower_tail (Bool, optional): Whether to return the lower tail of the distribution. Defaults to True.
        log_prob (Bool, optional): Whether to return the logarithm of the probability. Defaults to False.
        dtype (jnp.float32, optional): The data type of the output array. Defaults to jnp.float32.

    Returns:
        ArrayLike: An array containing the random binomial samples.
    """
    rvs = _rbinom(key, n, prob, sample_shape, dtype)
    if not lower_tail:
        rvs = 1 - rvs
    if log_prob:
        rvs = jnp.log(rvs)
    return rvs



import functools as ft

import jax.numpy as jnp
from jax import grad, jit, vmap


def squared_difference(x, y):
    return jnp.square(x - y)


def dtriangular(x, low, high, peak):
    x = jnp.asarray(x)
    x = jnp.atleast_1d(x)
    _dΔ = grad(_ptriangular)
    grads = vmap(_dΔ, in_axes=(0, None, None, None))(x, low, high, peak)
    return grads


def ptriangular(x, low, high, peak):
    x = jnp.asarray(x)
    x = jnp.atleast_1d(x)
    p = vmap(_ptriangular, in_axes=(0, None, None, None))(x, low, high, peak)
    return p


@ft.partial(
    jit,
    static_argnames=(
        "low",
        "high",
        "peak",
    ),
)
def _ptriangular(x, low, high, peak):
    interval_length = high - low

    cond0 = (x >= low) & (x <= peak)
    result_inside_interval = jnp.where(
        cond0,
        squared_difference(x, low) / (interval_length * (peak - low)),
        1.0 - squared_difference(high, x) / (interval_length * (high - peak)),
    )

    cond1 = x < low
    result_if_not_big = jnp.where(cond1, jnp.zeros_like(x), result_inside_interval)

    cond2 = x >= high
    return jnp.where(cond2, jnp.ones_like(x), result_if_not_big)



from functools import partial

import jax.numpy as jnp
import jax.tree_util as jtu
from equinox import filter_jit
from jax import Array, jit

from ...core import make_partial_pipe
from ...distribution import pF
from ...hypothesis import HypoTest


class OneWayAnovaTest(HypoTest):
    df_between: int
    df_within: int
    df_tol: int
    ss_between: float
    ss_within: float
    ss_total: float
    ms_between: float
    ms_within: float

    def __init__(
        self,
        statistic=None,
        p_value=None,
        df_between=None,
        df_within=None,
        df_total=None,
        ss_between=None,
        ss_within=None,
        ss_total=None,
        ms_between=None,
        ms_within=None,
        name="One Way Anova Test",
    ):
        super().__init__(statistic=statistic, p_value=p_value, name=name)
        self.df_between = df_between
        self.df_within = df_within
        self.df_tol = df_total
        self.ss_between = ss_between
        self.ss_within = ss_within
        self.ss_total = ss_total
        self.ms_between = ms_between
        self.ms_within = ms_within

    def _summary(self):
        template = """
                    ------------------------------
                    One-Way ANOVA Test Report
                    ------------------------------

                    Summary Statistics:
                    Number of samples analyzed: {}

                    Hypothesis Test:
                    Null Hypothesis (H0): "Groups are equal."
                    Alternative Hypothesis (H1): "Groups are not equal."

                    ANOVA Table:
                     Source    DF     SS         MS         F        P-value 
                    ---------------------------------------------------------
                     Between  {}     {:.3f}     {:.3f}     {:.3f}   {:.3f}   
                     Within   {}     {:.3f}     {:.3f}                       
                     Total    {}     {:.3f}                                    
                    """

        return template.format(
            self.df_tol + 1,
            self.df_between,
            self.ss_between,
            self.ms_between,
            self.statistic,
            self.p_value,
            self.df_within,
            self.ss_within,
            self.ms_within,
            self.df_tol,
            self.ss_total,
        )


@partial(jit, static_argnames=("axis"))
def _square_of_sums(a, axis=0) -> Array:
    s = jnp.sum(a, axis, keepdims=True)
    return s * s


@partial(jit, static_argnames=("axis"))
def _sum_of_squares(a, axis=0) -> Array:
    return jnp.sum(a * a, axis=axis, keepdims=True)


@make_partial_pipe
def one_way(*samples, axis=0) -> OneWayAnovaTest:
    """Performs a one-way ANOVA test.

    Args:
        *samples: A sequence of samples to compare.
        axis (int): The axis along which the samples are compared.

    Returns:
        OneWayAnovaTest: The result of the one-way ANOVA test.
    """
    samples = [jnp.asarray(sample) for sample in samples]
    ngroups = len(samples)
    return _one_way(samples, ngroups, axis=axis)


@filter_jit
def _one_way(samples, ngroups, axis=0):
    alldata = jnp.concatenate(samples, axis=axis)
    N = alldata.shape[axis]
    offset = jnp.mean(alldata, axis=axis, keepdims=True)
    alldata -= offset
    normalized_ss = _square_of_sums(alldata, axis=axis) / N

    # total
    sstot = _sum_of_squares(alldata, axis=axis) - normalized_ss

    # between groups or namely treatments
    ssbn = jnp.sum(
        jnp.concatenate(
            jtu.tree_map(
                lambda s: _square_of_sums(s - offset, axis) / s.shape[0], samples
            ),
            axis,
        ),
        axis,
        keepdims=True,
    )
    ssbn -= normalized_ss

    # with in groups
    sswn = sstot - ssbn

    # degree of freedom
    dfbn = ngroups - 1
    dfwn = N - ngroups

    # mean squares
    msb = ssbn / dfbn
    msw = sswn / dfwn

    # F-ratio
    f = msb / msw
    prob = 1 - pF(f, dfbn, dfwn)
    return OneWayAnovaTest(
        statistic=f[0],
        p_value=prob[0],
        df_between=dfbn,
        df_within=dfwn,
        df_total=N - 1,
        ss_between=ssbn[0],
        ss_within=sswn[0],
        ss_total=sstot[0],
        ms_between=msb[0],
        ms_within=msw[0],
    )



"""Test Basic Functions"""
import numpy as np
from absl.testing import absltest
from jax._src import test_util as jtest

from stamox.basic import mean, median, scale, sd, var
from stamox.core import Pipeable


class BasicTest(jtest.JaxTestCase):
    def test_mean(self):
        a = np.random.normal(size=(10000, 3))
        self.assertAllClose(mean(a, axis=1), np.mean(a, axis=1))

    def test_pipe_mean(self):
        a = np.random.normal(size=(10000, 3))
        p = Pipeable(a) >> mean(axis=1)
        self.assertAllClose(p(a), np.mean(a, axis=1))

    def test_median(self):
        a = np.random.normal(size=(10000, 3))
        self.assertAllClose(median(a, axis=1), np.median(a, axis=1))

    def test_pipe_median(self):
        a = np.random.normal(size=(10000, 3))
        p = Pipeable(a) >> median(axis=1)
        self.assertAllClose(p(a), np.median(a, axis=1))

    def test_std(self):
        a = np.random.normal(size=(10000, 3))
        self.assertAllClose(sd(a, axis=1), np.std(a, axis=1))

    def test_pipe_std(self):
        a = np.random.normal(size=(10000, 3))
        p = Pipeable(a) >> sd(axis=1)
        self.assertAllClose(p(a), np.std(a, axis=1))

    def test_var(self):
        a = np.random.normal(size=(10000, 3))
        self.assertAllClose(var(a, axis=1), np.var(a, axis=1))

    def test_pipe_var(self):
        a = np.random.normal(size=(10000, 3))
        p = Pipeable(a) >> var(axis=1)
        self.assertAllClose(p(a), np.var(a, axis=1))

    def test_scale(self):
        a = np.random.normal(size=(10000, 3)) * 2
        self.assertAllClose(
            scale(a), (a - np.mean(a, axis=0)) / np.std(a, axis=0, ddof=1)
        )


if __name__ == "__main__":
    absltest.main(testLoader=jtest.JaxTestLoader())



import functools as ft

import jax.numpy as jnp
from jax import jit

from ..distribution import pnorm, qnorm


def p_test(x, mu=0, alternative="two.sided", conf_level=0.95):
    return _p_test_single(x, mu, alternative, conf_level)

@ft.partial(jit, static_argnames=('alternative', 'mu', 'conf_level',))
def _p_test_single(x, mu=0, alternative="two.sided", conf_level=0.95):
    nx = x.shape[0]
    mx = jnp.mean(x, axis=-1, keepdims=True)
    vx = jnp.var(x, axis=-1, keepdims=True)
    stderr = jnp.sqrt(vx/nx)
    z_stat = (mx - mu) / stderr
    pval = 2 * pnorm(-jnp.abs(z_stat))
    alpha = jnp.array([1. - conf_level], dtype=x.dtype)
    conf_int  = qnorm(1 - alpha / 2)
    conf_int = jnp.array([-conf_int, conf_int]) + z_stat
    conf_int = conf_int * stderr + mu

    return z_stat, pval, conf_int


import jax.numpy as jnp
from equinox import filter_jit
from jax import lax, vmap
from jaxtyping import ArrayLike

from ..distribution import pnorm, qnorm
from ._base import HypoTest


_g = [-2.273, 0.459]
_c1 = [0.0, 0.221157, -0.147981, -2.07119, 4.434685, -2.706056]
_c2 = [0.0, 0.042981, -0.293762, -1.752461, 5.682633, -3.582633]
_c3 = [0.544, -0.39978, 0.025054, -6.714e-4]
_c4 = [1.3822, -0.77857, 0.062767, -0.0020322]
_c5 = [-1.5861, -0.31082, -0.083751, 0.0038915]
_c6 = [-0.4803, -0.082676, 0.0030302]
_a_1 = 0.70710678  # sqrt(2)


class ShapiroWilkTest(HypoTest):
    """Class to perform the Shapiro-Wilk test.

    This class is a subclass of HypoTest and provides methods to perform the Shapiro-Wilk test.

    Attributes:
        statistic (float): The test statistic.
        p_value (float): The p-value of the test.
    """

    def __init__(
        self,
        statistic=None,
        parameters=None,
        p_value=None,
        estimate=None,
        null_value=None,
        alternative=None,
        name="Shapiro-Wilk Test",
    ):
        super().__init__(
            statistic, parameters, p_value, estimate, null_value, alternative, name
        )

    def __repr__(self):
        return f"{self.name}(statistic={self.statistic}, parameters={self.parameters}, p_value={self.p_value})"


def shapiro_wilk_test_fun(x: ArrayLike) -> ShapiroWilkTest:
    """Computes the Shapiro-Wilk test for normality.

    Args:
        x (ArrayLike): The data to be tested.

    Returns:
        ShapiroWilkTest: The Shapiro-Wilk Test object.

    Example:
        >>> import jax.numpy as jnp
        >>> from stamox.hypothesis import shapiro_wilk_test
        >>> x = jnp.array([1, 2, 3, 4, 5])
        >>> shapiro_wilk_test(x)
        Shapiro-Wilk Test(statistic=0.0, p_value=None)
    """
    x = jnp.ravel(x)
    x = jnp.sort(x)
    n = x.size
    x = x - jnp.median(x)
    w = _shapiro_wilk(x, n)
    return w


@filter_jit
def _shapiro_wilk(x, n):
    nn2 = n // 2
    an = jnp.asarray(n, dtype=jnp.float32)
    if n < 3:
        raise ValueError("Data must be at least length 3.")

    an25 = an + 0.25
    a = vmap(lambda i: qnorm(jnp.asarray((i - 0.375) / an25), 0.0, 1.0), in_axes=0)(
        jnp.arange(nn2, dtype=jnp.float32) + 1
    )
    a = jnp.squeeze(a, axis=-1)
    summ2 = 2 * jnp.sum(a**2, axis=0)
    ssumm2 = jnp.sqrt(summ2)
    rsn = 1.0 / jnp.sqrt(an)
    a1 = poly(rsn, _c1, 6) - a[0] / ssumm2

    # Normalize
    if n > 5:
        a2 = -a[1] / ssumm2 + poly(rsn, _c2, 6)
        fac = jnp.sqrt(
            (summ2 - 2.0 * (a[0] * a[0]) - 2.0 * (a[1] * a[1]))
            / (1.0 - 2.0 * (a1 * a1) - 2.0 * (a2 * a2))
        )
    else:
        fac = jnp.sqrt((summ2 - 2.0 * (a[0] * a[0])) / (1.0 - 2.0 * (a1 * a1)))
    if n > 3:
        a2 = a[1] / -fac
        aa = jnp.r_[a1, a2, jnp.zeros((n // 2 - 2,))]
        aaa = jnp.r_[0.0, 0.0, a[2:]] / -fac
        a = aaa + aa
    else:
        a = jnp.r_[a1]

    range_x = x[n - 1] - x[0]
    sx = jnp.sum(x, axis=0) / range_x

    forward = jnp.arange(1, n, 1, dtype=jnp.float32)
    reverse = jnp.arange(n - 1, 0, -1, dtype=jnp.float32) - 1
    sa = jnp.sum(
        vmap(
            lambda i, j: jnp.sign(i - j) * a[jnp.int32(jnp.minimum(i, j))],
            in_axes=(0, 0),
        )(forward, reverse),
        axis=0,
    )
    sa += -a[0]
    sa /= n
    sx /= n

    def _tmp(i, j, xi):
        asa = jnp.sign(i - j) * a[jnp.int32(jnp.minimum(i, j))] - sa
        xsx = xi / range_x - sx
        ssa = asa * asa
        ssx = xsx * xsx
        sax = asa * xsx
        return (ssa, ssx, sax)

    forward = jnp.arange(0, n, 1, dtype=jnp.float32)
    reverse = jnp.arange(n, 0, -1, dtype=jnp.float32) - 1
    (ssa, ssx, sax) = vmap(_tmp, in_axes=(0, 0, 0))(forward, reverse, x)
    ssa = jnp.sum(ssa, axis=0, keepdims=True)
    ssx = jnp.sum(ssx, axis=0, keepdims=True)
    sax = jnp.sum(sax, axis=0, keepdims=True)

    ssassx = jnp.sqrt(ssa * ssx)
    w1 = (ssassx - sax) * (ssassx + sax) / (ssa * ssx)
    w = 1.0 - w1

    # cal pw
    if n == 3:
        pi6 = 1.90985931710274
        stqr = 1.04719755119660
        pw = pi6 * (jnp.arcsin(jnp.sqrt(w)) - stqr)
        return ShapiroWilkTest(statistic=w, p_value=pw)

    y = jnp.log(w1)
    xx = jnp.log(an)
    if n <= 11:
        gamma = poly(an, _g, 2)

        pw = lax.select(
            y >= gamma,
            jnp.array([0.0]),
            pnorm(
                -jnp.log(gamma - y),
                poly(an, _c3, 4),
                jnp.exp(poly(an, _c4, 4)),
                lower_tail=False,
            ),
        )
    else:
        m = poly(xx, _c5, 4)
        s = jnp.exp(poly(xx, _c6, 3))
        pw = pnorm(y, m, s, lower_tail=False).squeeze()
    w = w.squeeze()
    return ShapiroWilkTest(statistic=w, p_value=pw)


@filter_jit
def poly(x, coef, norder):
    coef = jnp.asarray(coef)
    iss = jnp.arange(norder, dtype=jnp.float32)
    value = vmap(lambda c, i, x: x**i * c, in_axes=(0, 0, None))(coef, iss, x)
    return jnp.sum(value, axis=0, keepdims=True)



from typing import Optional, Union

import jax.numpy as jnp
import jax.random as jrand
from equinox import filter_grad, filter_jit, filter_vmap
from jax._src.random import KeyArray, Shape
from jaxtyping import ArrayLike, Float

from ..core import make_partial_pipe


@filter_jit
def _pweibull(
    x: Union[Float, ArrayLike],
    concentration: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
):
    return 1 - jnp.exp(-((x / scale) ** concentration))


@make_partial_pipe
def pweibull(
    q: Union[Float, ArrayLike],
    concentration: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
    lower_tail: bool = True,
    log_prob: bool = False,
) -> ArrayLike:
    """Computes the cumulative distribution function of the Weibull distribution.

    Args:
        q (Union[Float, ArrayLike]): The value at which to evaluate the CDF.
        concentration (Union[Float, ArrayLike], optional): The concentration parameter of the Weibull distribution. Defaults to 0.0.
        scale (Union[Float, ArrayLike], optional): The scale parameter of the Weibull distribution. Defaults to 1.0.
        lower_tail (bool, optional): Whether to compute the lower tail of the CDF. Defaults to True.
        log_prob (bool, optional): Whether to return the log probability. Defaults to False.

    Returns:
        Array: The cumulative distribution function of the Weibull distribution evaluated at `q`.
    """
    q = jnp.asarray(q)
    q = jnp.atleast_1d(q)
    p = filter_vmap(_pweibull)(q, concentration, scale)
    if not lower_tail:
        p = 1 - p
    if log_prob:
        p = jnp.log(p)
    return p


_dweibull = filter_grad(filter_jit(_pweibull))


@make_partial_pipe
def dweibull(
    x, concentration=0.0, scale=1.0, lower_tail=True, log_prob=False
) -> ArrayLike:
    """Computes the probability density function of the Weibull distribution.

    Args:
        x (Union[Float, ArrayLike]): The value at which to evaluate the PDF.
        concentration (Union[Float, ArrayLike], optional): The concentration parameter of the Weibull distribution. Defaults to 0.0.
        scale (Union[Float, ArrayLike], optional): The scale parameter of the Weibull distribution. Defaults to 1.0.
        lower_tail (bool, optional): Whether to compute the lower tail of the CDF. Defaults to True.
        log_prob (bool, optional): Whether to return the log probability. Defaults to False.

    Returns:
        Array: The probability density function of the Weibull distribution evaluated at `x`.
    """
    x = jnp.asarray(x)
    x = jnp.atleast_1d(x)
    grads = filter_vmap(_dweibull)(x, concentration, scale)
    if not lower_tail:
        grads = 1 - grads
    if log_prob:
        grads = jnp.log(grads)
    return grads


@filter_jit
def _qweibull(
    q: Union[Float, ArrayLike],
    concentration: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
) -> ArrayLike:
    x = jnp.float_power(-jnp.log(1 - q), 1 / concentration) * scale
    return x


@make_partial_pipe
def qweibull(
    p: Union[Float, ArrayLike],
    concentration: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
    lower_tail: bool = True,
    log_prob: bool = False,
) -> ArrayLike:
    """Computes the quantile function of the Weibull distribution.

    Args:
        p (Union[Float, ArrayLike]): The quantiles to compute.
        concentration (Union[Float, ArrayLike], optional): The concentration parameter of the Weibull distribution. Defaults to 0.0.
        scale (Union[Float, ArrayLike], optional): The scale parameter of the Weibull distribution. Defaults to 1.0.
        lower_tail (bool, optional): Whether to compute the lower tail of the distribution. Defaults to True.
        log_prob (bool, optional): Whether to compute the log probability of the distribution. Defaults to False.

    Returns:
        Array: The computed quantiles.
    """
    p = jnp.asarray(p)
    p = jnp.atleast_1d(p)
    if not lower_tail:
        p = 1 - p
    if log_prob:
        p = jnp.exp(p)
    q = filter_vmap(_qweibull)(p, concentration, scale)
    return q


@filter_jit
def _rweibull(
    key: KeyArray,
    concentration: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
    sample_shape: Optional[Shape] = None,
):
    return jrand.weibull_min(key, scale, concentration, sample_shape)


@make_partial_pipe
def rweibull(
    key: KeyArray,
    sample_shape: Optional[Shape] = None,
    concentration: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
    lower_tail: bool = True,
    log_prob: bool = False,
) -> ArrayLike:
    """Generates samples from the Weibull distribution.

    Args:
        key (KeyArray): Random key used for generating random numbers.
        sample_shape (Optional[Shape], optional): Shape of the output sample. Defaults to None.
        concentration (Union[Float, ArrayLike], optional): Concentration parameter of the Weibull distribution. Defaults to 0.0.
        scale (Union[Float, ArrayLike], optional): Scale parameter of the Weibull distribution. Defaults to 1.0.
        lower_tail (bool, optional): Whether to return the lower tail probability. Defaults to True.
        log_prob (bool, optional): Whether to return the log probability. Defaults to False.

    Returns:
        rvs (ArrayLike): Probability of the Weibull distribution.
    """
    rvs = _rweibull(key, concentration, scale, sample_shape)
    if not lower_tail:
        rvs = 1 - rvs
    if log_prob:
        rvs = jnp.log(rvs)
    return rvs



from stamox.core import make_pipe
from stamox.sample._bootstrap import bootstrap, bootstrap_sample
from stamox.sample._jackknife import jackknife, jackknife_sample_fun


jackknife_sample = make_pipe(jackknife_sample_fun, name="jackknife_sample")

__all__ = [
    "bootstrap",
    "bootstrap_sample",
    "jackknife",
    "jackknife_sample",
]



from stamox.distribution._beta import dbeta, pbeta, qbeta, rbeta
from stamox.distribution._binomial import dbinom, pbinom, qbinom, rbinom
from stamox.distribution._cauchy import dcauchy, pcauchy, qcauchy, rcauchy
from stamox.distribution._chisq import dchisq, pchisq, qchisq, rchisq
from stamox.distribution._ecdf import ecdf, step_fun
from stamox.distribution._exp import dexp, pexp, qexp, rexp
from stamox.distribution._f import dF, pF, qF, rF
from stamox.distribution._gamma import dgamma, pgamma, qgamma, rgamma
from stamox.distribution._laplace import dlaplace, plaplace, qlaplace, rlaplace
from stamox.distribution._normal import dnorm, pnorm, qnorm, rnorm
from stamox.distribution._pareto import dpareto, ppareto, qpareto, rpareto
from stamox.distribution._poisson import dpoisson, ppoisson, qpoisson, rpoisson
from stamox.distribution._t import dt, pt, qt, rt
from stamox.distribution._uniform import dunif, punif, qunif, runif


__all__ = [
    "pt",
    "qt",
    "rt",
    "dt",
    "pbinom",
    "qbinom",
    "rbinom",
    "dbinom",
    "pnorm",
    "qnorm",
    "rnorm",
    "dnorm",
    "pbeta",
    "qbeta",
    "rbeta",
    "dbeta",
    "pgamma",
    "qgamma",
    "rgamma",
    "dgamma",
    "runif",
    "punif",
    "qunif",
    "dunif",
    "pchisq",
    "qchisq",
    "rchisq",
    "dchisq",
    "ppareto",
    "qpareto",
    "rpareto",
    "dpareto",
    "ppoisson",
    "qpoisson",
    "rpoisson",
    "dpoisson",
    "pF",
    "qF",
    "dF",
    "rF",
    "pcauchy",
    "qcauchy",
    "rcauchy",
    "dcauchy",
    "pexp",
    "qexp",
    "rexp",
    "dexp",
    "plaplace",
    "qlaplace",
    "dlaplace",
    "rlaplace",
    "ecdf",
    "step_fun",
]



from typing import Callable

import jax.numpy as jnp
from jax import jit
from jaxtyping import ArrayLike

from ..core import make_partial_pipe


def step_fun(x, y, ival=0.0, sorted=False, side="left", dtype=jnp.float32):
    """Returns a function that evaluates a step function at given points.

    Args:
        x (array-like): The x-coordinates of the step points.
        y (array-like): The y-coordinates of the step points.
        ival (float, optional): The initial value of the step function. Defaults to 0.
        sorted (bool, optional): Whether the x-coordinates are already sorted. Defaults to False.
        side (str, optional): The side of the interval to take when evaluating the step function. Must be either 'left' or 'right'. Defaults to 'left'.
        dtype (jnp.dtype, optional): The dtype of the output. Defaults to jnp.float32.

    Returns:
        Callable[..., ArrayLike]: A function that evaluates the step function at given points.
    """
    if side.lower() not in ["right", "left"]:
        raise ValueError("side must be left or right")

    _x = jnp.asarray(x, dtype=dtype)
    _y = jnp.asarray(y, dtype=dtype)

    _x = jnp.r_[-jnp.inf, _x]
    _y = jnp.r_[ival, _y]

    if not sorted:
        asort = jnp.argsort(_x)
        _x = jnp.take(_x, asort, 0)
        _y = jnp.take(_y, asort, 0)

    @jit
    def _call(time):
        time = jnp.asarray(time)
        tind = jnp.searchsorted(_x, time, side) - 1
        return _y[tind]

    return _call


@make_partial_pipe
def ecdf(x: ArrayLike, side="right", dtype=jnp.float32) -> Callable[..., ArrayLike]:
    """Calculates the empirical cumulative distribution function (ECDF) of a given array.

    Args:
        x (array): The array to calculate the ECDF for.
        side (str, optional): Specifies which side of the step function to use. Defaults to 'right'.
        dtype (jnp.dtype, optional): The dtype of the output. Defaults to jnp.float32.

    Returns:
        Callable[..., ArrayLike]: A function that evaluates the ECDF at given points.
    """
    x = jnp.array(x, copy=True, dtype=dtype)
    x = jnp.sort(x)
    nobs = x.size
    y = jnp.linspace(1.0 / nobs, 1, nobs)
    return step_fun(x, y, side=side, sorted=True)



import jax.numpy as jnp
from equinox import filter_jit
from jax import vmap
from jaxtyping import ArrayLike

from ...core import Functional, StateFunc


class PCAState(StateFunc):
    n_components: int
    components: ArrayLike
    mean: ArrayLike

    def __init__(self, n_components, components, mean):
        super().__init__(name="PCAState", fn=None)
        self.n_components = n_components
        self.components = components
        self.mean = mean
    


class PCA(Functional):
    n_components: int

    def __init__(self, n_components):
        super().__init__(name="PCA", fn=None)
        self.n_components = n_components

    def __call__(self, X, *args, **kwargs) -> PCAState:
        return self._pca(X)

    @filter_jit
    def _pca(self, X):
        # Subtract the mean of each feature column to center the data around the origin
        _mean = jnp.mean(X, axis=0)
        X_meaned = vmap(lambda x: x - _mean)(X)
        # Computing the covariance matrix
        cov_mat = jnp.cov(X_meaned.T)
        
        # Computing the eigenvalues and eigenvectors of the covariance matrix
        eigen_vals, eigen_vecs = jnp.linalg.eigh(cov_mat)
        
        # Sorting the eigenvalues in descending order and selecting the top n_components eigenvectors
        sorted_idx = eigen_vals.argsort()[::-1]
        sorted_eigen_vals = eigen_vals[sorted_idx]
        sorted_eigen_vecs = eigen_vecs[:, sorted_idx]
        eigenvectors_subset = sorted_eigen_vecs[:, :self.n_components]

        return PCAState(
            n_components=self.n_components, components=eigenvectors_subset, mean=_mean
        )



import pathlib

from setuptools import find_packages, setup


HERE = pathlib.Path(__file__).parent

VERSION = "0.1.2"
PACKAGE_NAME = "stamox"
AUTHOR = "Jia Yaobo"
URL = "https://github.com/jiayaobo/stamox"

LICENSE = "Apache 2.0"
DESCRIPTION = "Accelerate Your Statistical Analysis with JAX."
LONG_DESCRIPTION = (HERE / "README.md").read_text()
LONG_DESC_TYPE = "text/markdown"

INSTALL_REQUIRES = [
    "jax>=0.4.6",
    "jaxtyping>=0.2.14",
    "typing_extensions>=4.5.0",
    "equinox>=0.10.1",
    "jaxopt>=0.6",
    "pandas>=1.5.3",
    "patsy>=0.5.3",
    "tensorflow-probability>=0.19.0",
    "scipy"
]
TESTS_REQUIRES = ["pytest", "scipy", "numpy", "sklearn", "statsmodels"]

setup(
    name=PACKAGE_NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type=LONG_DESC_TYPE,
    author=AUTHOR,
    license=LICENSE,
    url=URL,
    install_requires=INSTALL_REQUIRES,
    extras_require={
        "tests": TESTS_REQUIRES,
        "complete": INSTALL_REQUIRES + TESTS_REQUIRES,
    },
    packages=find_packages(),
    python_requires=">=3.8",
)



import jax.numpy as jnp
from jax import vmap


def align_inputs(*inputs):

    max_size = 1
    n = len(inputs)

    args = []
    arg_sizes = []

    if n == 1:
        arg = jnp.asarray(inputs[0]).reshape(-1)
        arg = jnp.atleast_1d(inputs[0])
        return arg, [arg.size]

    for arg in inputs:
        arg = jnp.asarray(arg)

        if arg.size > 1:
            arg = arg.reshape(-1)

        arg_sizes.append(arg.size)
        args.append(arg)
        if arg.size > max_size:
            max_size = arg.size

    for i in range(n):
        if not (args[i].size == 1 or args[i].size == max_size):
            args[i] = jnp.tile(
                args[i], 1+(max_size // args[i].size))[:max_size]
            arg_sizes[i] = max_size

    return args, arg_sizes


def auto_map(func, *inputs):

    n = len(inputs)

    args, arg_sizes = align_inputs(*inputs)

    in_axes = []
    all_scalar = True
    for arg_size in arg_sizes:
        if arg_size > 1:
            all_scalar = False
            in_axes.append(0)
        else:
            in_axes.append(None)
    
    if all_scalar:
        return jnp.atleast_1d(func(*args))

    if n == 1:
        return vmap(func, in_axes=0)(args)
        
    return vmap(func, in_axes=in_axes)(*args)



from stamox.core import make_pipe
from stamox.hypothesis._bartlett import bartlett_test_fun, BartlettTest
from stamox.hypothesis._base import HypoTest
from stamox.hypothesis._durbin_watson import durbin_watson_test, DurbinWatsonTest
from stamox.hypothesis._friedman import friedman_test_fun, FriedmanTest
from stamox.hypothesis._shapiro_wilk import shapiro_wilk_test_fun, ShapiroWilkTest


bartlett_test = make_pipe(bartlett_test_fun, name="bartlett_test")
friedman_test = make_pipe(friedman_test_fun, name="friedman_test")
shapiro_wilk_test = make_pipe(shapiro_wilk_test_fun, name="shapiro_wilk_test")

__all__ = [
    "HypoTest",
    "bartlett_test",
    "BartlettTest",
    "durbin_watson_test",
    "DurbinWatsonTest",
    "friedman_test",
    "FriedmanTest",
    "shapiro_wilk_test",
    "ShapiroWilkTest",
]



import jax.numpy as jnp
from equinox import filter_jit
from jaxtyping import ArrayLike

from ..core import make_partial_pipe
from ._base import HypoTest


class DurbinWatsonTest(HypoTest):
    """
    Class for performing the Durbin-Watson Test.

    This class is a subclass of HypoTest and provides methods to perform the Durbin-Waston Test.

    Attributes:
        statistic (float): The test statistic.
        parameters (tuple): The parameters of the test.
        p_value (float): The p-value of the test.
    """

    def __init__(
        self,
        statistic=None,
        parameters=None,
        p_value=None,
        estimate=None,
        null_value=None,
        alternative=None,
        name="Durbin-Watson Test",
    ):
        super().__init__(
            statistic, parameters, p_value, estimate, null_value, alternative, name
        )

    def __repr__(self):
        return f"{self.name}(statistic={self.statistic}, parameters={self.parameters}, p_value={self.p_value})"


@make_partial_pipe
def durbin_watson_test(resids: ArrayLike, axis: int = 0):
    """Computes the Durbin-Watson statistic for a given array of residuals.

    Args:
        resids (array): An array of residuals.
        axis (int, optional): The axis along which to compute the statistic. Defaults to 0.

    Returns:
        DurbinWatsonTest: The Durbin-Watson Test object.

    Example:
        >>> import jax.numpy as jnp
        >>> from stamox.hypothesis import durbin_watson_test
        >>> resids = jnp.array([1, 2, 3, 4, 5])
        >>> durbin_watson_test(resids)
        Durbin-Waston Test(statistic=0.0, parameters=None, p_value=None)
    """
    resids = jnp.atleast_1d(resids)
    return _durbin_watson(resids, axis)


@filter_jit
def _durbin_watson(resids, axis=0):
    """Computes the Durbin-Watson statistic for a given array of residuals.

    Args:
        resids (array): An array of residuals.
        axis (int, optional): The axis along which to compute the statistic. Defaults to 0.

    Returns:
        float: The Durbin-Watson statistic.
    """
    diff_resids = jnp.diff(resids, 1, axis=axis)
    dw = jnp.sum(diff_resids**2, axis=axis, keepdims=True) / jnp.sum(
        resids**2, axis=axis, keepdims=True
    )
    return DurbinWatsonTest(statistic=dw)



import functools as ft

import jax.numpy as jnp
from jax import jit

from stamox.distribution import pt, qt


def t_test(x, mu=0, alternative="two.sided", conf_level=0.95):
    return _t_test_single(x, mu, alternative, conf_level)


@ft.partial(
    jit,
    static_argnames=(
        "alternative",
        "mu",
        "conf_level",
    ),
)
def _t_test_single(x, mu=0, alternative="two.sided", conf_level=0.95):
    nx = x.shape[0]
    mx = jnp.mean(x, axis=-1, keepdims=True)
    vx = jnp.var(x, axis=-1, keepdims=True)

    df = nx - 1
    stderr = jnp.sqrt(vx / nx)
    t_stat = (mx - mu) / stderr
    pval = 2 * pt(-jnp.abs(t_stat), df)
    alpha = jnp.array([1.0 - conf_level], dtype=x.dtype)
    conf_int = qt(1 - alpha / 2, df)
    conf_int = jnp.array([-conf_int, conf_int]) + t_stat
    conf_int = conf_int * stderr + mu

    return t_stat, pval, conf_int


def _t_test_pair(x, y, alternativate="two.sided", mu=0, conf_level=0.95):
    pass



import functools as ft

import jax.numpy as jnp
import jax.scipy as jsp
import jax.tree_util as jtu
from jax import jit, vmap

import numpy as np

from ..core import Model


class Corr(Model):

    def __init__(self) -> None:
        super().__init__()

    def __call__(self, x1, x2, tolerance=1e-8):
        _corr(x1, x2, tolerance=tolerance)


@ft.partial(jit, static_argnames=('tolerance', ))
def _corr(x1, x2, tolerance=1e-8):
    nobs, k_yvar = x1.shape
    nobs, k_xvar = x2.shape

    k = np.min([k_yvar, k_xvar])

    x = jnp.array(x1)
    y = jnp.array(x2)

    x = x - x.mean()
    y = y - y.mean()

    ux, sx, vx = jsp.linalg.svd(x)
    # vx_ds = vx.T divided by sx
    vx_ds = vx.T
    uy, sy, vy = jsp.linalg.svd(y)
    # vy_ds = vy.T divided by sy
    vy_ds = vy.T
    u, s, v = jsp.linalg.svd(ux.T.dot(uy))

    # Correct any roundoff
    corr = vmap(lambda c: jnp.maximum(0, jnp.minimum(c, 1)), in_axes=0)(s)

    x_coef = vx_ds.dot(u[:, :k])
    y_coef = vy_ds.dot(v.T[:, :k])

    return corr, x_coef, y_coef



from typing import Optional, Union

import jax.numpy as jnp
import jax.random as jrand
import tensorflow_probability.substrates.jax.math as tfp_math
from equinox import filter_grad, filter_jit, filter_vmap
from jax._src.random import KeyArray, Shape
from jax.scipy.special import gammainc
from jaxtyping import ArrayLike, Float

from ..core import make_partial_pipe


@filter_jit
def _pgamma(
    q, shape: Union[Float, ArrayLike] = 1.0, rate: Union[Float, ArrayLike] = 1.0
):
    return gammainc(shape, q * rate)


@make_partial_pipe
def pgamma(
    q: Union[Float, ArrayLike],
    shape: Union[Float, ArrayLike] = 1.0,
    rate: Union[Float, ArrayLike] = 1.0,
    lower_tail: bool = True,
    log_prob: bool = False,
    dtype=jnp.float32,
) -> ArrayLike:
    """Computes the cumulative distribution function of the gamma distribution.

    Args:
        q: A float or array-like object representing the input to the gamma function.
        shape: A float or array-like object representing the shape parameter of the gamma function.
        rate: A float or array-like object representing the rate parameter of the gamma function.
        lower_tail: A boolean indicating whether to compute the lower tail of the gamma function.
        log_prob: A boolean indicating whether to compute the logarithm of the probability density function.
        dtype: The dtype of the output. Defaults to float32.

    Returns:
        ArrayLike: The CDF value of the given value or array of values.

    Example:
        >>> pgamma(1.0, 0.5, 0.5)
        Array([0.6826893], dtype=float32, weak_type=True)
    """
    q = jnp.asarray(q, dtype=dtype)
    q = jnp.atleast_1d(q)
    p = filter_vmap(_pgamma)(q, shape, rate)
    if not lower_tail:
        p = 1.0 - p
    if log_prob:
        p = jnp.log(p)
    return p


_dgamma = filter_jit(filter_grad(_pgamma))


@make_partial_pipe
def dgamma(
    x: Union[Float, ArrayLike],
    shape: Union[Float, ArrayLike] = 1.0,
    rate: Union[Float, ArrayLike] = 1.0,
    lower_tail: bool = True,
    log_prob: bool = False,
    dtype=jnp.float32,
) -> ArrayLike:
    """Compute density of gamma distribution.

    Args:
        x (Union[Float, ArrayLike]): The value at which to evaluate the gamma
            distribution.
        shape (Union[Float, ArrayLike], optional): The shape parameter of the
            gamma distribution. Defaults to 1.0.
        rate (Union[Float, ArrayLike], optional): The rate parameter of the
            gamma distribution. Defaults to 1.0.
        lower_tail (bool, optional): Whether to compute the lower tail of the
            gamma distribution. Defaults to True.
        log_prob (bool, optional): Whether to return the log probability.
            Defaults to False.
        dtype (jnp.dtype, optional): The dtype of the output. Defaults to

    Returns:
        ArrayLike: The density of the gamma distribution evaluated
            at `x`. If `log_prob` is True, returns the log probability.

    Example:
        >>> dgamma(1.0, 0.5, 0.5)
        Array([0.24197064], dtype=float32, weak_type=True)
    """
    x = jnp.asarray(x, dtype=dtype)
    x = jnp.atleast_1d(x)
    grads = filter_vmap(_dgamma)(x, shape, rate)
    if not lower_tail:
        grads = 1 - grads
    if log_prob:
        grads = jnp.log(grads)
    return grads


@filter_jit
def _qgamma(
    q: Union[Float, ArrayLike],
    shape: Union[Float, ArrayLike] = 1.0,
    rate: Union[Float, ArrayLike] = 1.0,
):
    return tfp_math.igammainv(shape, q) / rate


@make_partial_pipe
def qgamma(
    p: Union[Float, ArrayLike],
    shape: Union[Float, ArrayLike] = 1.0,
    rate: Union[Float, ArrayLike] = 1.0,
    lower_tail: bool = True,
    log_prob: bool = False,
    dtype=jnp.float32,
) -> ArrayLike:
    """Computes the quantile of the gamma distribution.

    Args:
        p: A float or array-like object representing the quantile.
        shape: A float or array-like object representing the shape parameter of the gamma distribution.
        rate: A float or array-like object representing the rate parameter of the gamma distribution.
        lower_tail: A boolean indicating whether to compute the lower tail (default) or upper tail.
        log_prob: A boolean indicating whether to compute the log probability (default False).
        dtype: The dtype of the output. Defaults to float32.

    Returns:
        ArrayLike: The quantile of the gamma distribution.

    Example:
        >>> qgamma(0.5, 0.5, 0.5)
        Array([0.45493677], dtype=float32)
    """
    p = jnp.asarray(p, dtype=dtype)
    p = jnp.atleast_1d(p)
    if not lower_tail:
        p = 1 - p
    if log_prob:
        p = jnp.exp(p)
    x = filter_vmap(_qgamma)(p, shape, rate)
    return x


@make_partial_pipe
def rgamma(
    key,
    sample_shape: Optional[Shape] = None,
    shape: Union[Float, ArrayLike] = 1.0,
    rate: Union[Float, ArrayLike] = 1.0,
    lower_tail: bool = True,
    log_prob: bool = False,
    dtype=jnp.float32,
) -> ArrayLike:
    """Generates random gamma values.

    Args:
        key: A PRNGKey to use for the random number generation.
        sample_shape: An optional shape for the output array.
        shape: The shape parameter of the gamma distribution.
        rate: The rate parameter of the gamma distribution.
        lower_tail: Whether to return the lower tail of the distribution.
        log_prob: Whether to return the log probability of the result.
        dtype: The dtype of the output. Defaults to float32.

    Returns:
        ArrayLike: A random gamma value or an array of random gamma values.

    Example:
        >>> rgamma(key, shape=0.5, rate=0.5)
        Array(0.3384059, dtype=float32)
    """
    rvs = _rgamma(key, shape, rate, sample_shape, dtype=dtype)
    if not lower_tail:
        rvs = 1 - rvs
    if log_prob:
        rvs = jnp.log(rvs)
    return rvs


@filter_jit
def _rgamma(
    key: KeyArray,
    shape: Union[Float, ArrayLike] = 1.0,
    rate: Union[Float, ArrayLike] = 1.0,
    sample_shape: Optional[Shape] = None,
    dtype=jnp.float32,
):
    if sample_shape is None:
        sample_shape = jnp.broadcast_shapes(jnp.shape(shape), jnp.shape(rate))
    shape = jnp.broadcast_to(shape, sample_shape)
    rate = jnp.broadcast_to(rate, sample_shape)
    return jrand.gamma(key, shape, sample_shape, dtype) / rate



import jax.numpy as jnp
from equinox import filter_jit
from jax import lax

from ..maps import auto_map


@filter_jit
def _cumm_dgeom(k ,p):

    def cond(carry):
        i, k,  _ = carry
        return i <= k

    def body(carry):
        i, k, ds0 = carry
        ds1 = ds0 + dgeom(i, p)
        i = i + 1
        carry = (i, k, ds1)
        return carry

    i = 0
    init = (i,  k,  jnp.asarray([0.]))
    out = lax.while_loop(cond, body, init)
    return out[2]

def pgeom(k ,p):
    k = jnp.asarray(k ,jnp.int32)
    pp = auto_map(_cumm_dgeom, k, p)
    pp = lax.clamp(0., pp, 1.).ravel()
    return pp



@filter_jit
def _dgeom(k, p):
    return jnp.power(1-p, k) * p

def dgeom(k, p):
    # k = jnp.asarray(k, dtype=jnp.int32)
    pp = auto_map(_dgeom, k, p)
    return pp





